// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef PPSSUMM_DEGREES
#define PPSSUMM_DEGREES
#include <cstdlib>
#include <cmath>
#include <cerrno>
#include "BaseReal.h"
#include "Dal.h"
#include <stdio.h>

class Degrees : public BaseReal<real64>
{
    public:
    enum Format 
    {
    DECIMALDEGREES
    ,DECIMALHOURS
    ,HMS
    ,DMS
    };

    Degrees() : BaseReal<real64>(0.0) { } 
    Degrees( const Degrees& s ) : BaseReal<real64>(s.get()) { } 
    Degrees( const real64 s ) : BaseReal<real64>(s) { } 
    Degrees( const  char* s ) : BaseReal<real64>( strtod( s , NULL ) ) { } 
    Degrees( const std::string& s ) : BaseReal<real64>( strtod( s.c_str() , NULL ) ) { } 
    Degrees( const real64 s , const Degrees::Format fmt ) : BaseReal<real64>(0.0) 
    {
    set_fromReal64(s,fmt);
    }
    Degrees( const std::string& s , const Degrees::Format fmt ) : BaseReal<real64>(0.0) 
    {
    set_fromString(s,fmt);
    }

    void set_fromReal64( const real64 s , const Degrees::Format fmt )
    {
    if (fmt == DECIMALDEGREES)
    {
        
        this->set( s ) ;
        _format = fmt;
    }
    else if (fmt == DECIMALHOURS)
    {
        this->set( s * 15.0 );
        _format = fmt;
    }
    } 


    void set_fromString( const std::string& s ,  const Degrees::Format fmt )
    {
    if (fmt == DECIMALDEGREES)
    {
        this->set( strtod( s.c_str() , NULL ) );
        _format = fmt;
    }
    else if (fmt == DECIMALHOURS)
    {

        real64 d = strtod( s.c_str() , NULL );
        this->set( d * 15.0 );
        _format = fmt;
    }
    else if (fmt == HMS)
    {
        // Not yet implemented
        _format = fmt;
    }
    else if (fmt == DMS)
    {
        // Not yet implemented
        _format = fmt;
    }
    } 

    /* Precission of output
     * When hms or dms output are selected this applies to the seconds field.
     * When decimal output is selected it applies to the real number written out.
     */
    int getprec() const { return _precision; };
    int setprec(const int p)  { int tmp = _precision; _precision = p ; return tmp; }

    Degrees operator=(const Degrees& s)
    {
    this->set( s.get() );
    return *this;
    }
    Degrees operator=(const char* s)
    {
    this->set( static_cast<real64>( strtod( s , NULL ) ) );
    return *this;
    }
    Degrees operator=(const std::string& s)
    {
    this->set( static_cast<real64>( strtod( s.c_str() , NULL ) ) );
    return *this;
    }

    /* integer degrees */
    int degrees() const
    {
	return static_cast<int>(this->get());
    }

    void asHms( int& hr_i , int& min_i , real64& sec ) const
    {
	/* Convert this angle to hh mm ss.ss and write these into the call references */
	real64 hr = fmod( ( this->get() + 360.0 ), 360.0 ) / 15 ;
	hr_i = static_cast<int>(hr);

	real64 min = hr * 60.0 - static_cast<real64>(hr_i) * 60.0;
	min_i = static_cast<int>(min);

	sec = min * 60.0 - static_cast<real64>(min_i) * 60.0;
    }

    std::string asHms() const
    {
    /* Return a string of this angle in  hhHmmMss.sssS format.
     * This has no bearing on the operator<< output
     */
        int hr=0 , min = 0;
        real64 sec = 0.0;
        this->asHms(hr , min , sec);
       
        int p = this->getprec();
        char c[255];

        sprintf( c , "%02d&nbsp;%02d&nbsp;%0*.*f" , hr , min , p+2 , p , sec );
        return std::string(c);
    }

    std::string asAbrevHms() const
    {
    /* Return a string of this angle in  ddDmmMss.sssS format */
	int hr=0 , min = 0;
        real64 sec = 0.0;
        this->asHms(hr , min , sec);
       
        int p = this->getprec();
        char c[255];
	if (hr)
	    sprintf( c , "%02d&nbsp;%02d&nbsp;%0*.*f" , hr , min , p+2 , p , sec );
	else if (min)
	    sprintf( c , "%02d&nbsp;%*.*f" , min , p+2 , p , sec );
	else 
	    sprintf( c , "%*.*f" , p+2 , p , sec );

        return std::string(c);    
    }



    void asDms( bool& negative , int& deg_i , int& min_i , real64& sec ) const 
    {
    /* Convert this angle to dd mm ss.ss and write these into the call references
     * This has no bearing on the operator<< output
     */
        negative = false;
	real64 deg = this->get();
        if ( deg  < 0 ) 
	{
	    negative = 1;
	    deg = fabs(deg);
	}
	deg_i = static_cast<int>(deg);
	
	real64 min = deg * 60.0 - static_cast<real64>(deg_i) * 60.0;
	min_i = static_cast<int>(min);

	sec = min * 60.0 - static_cast<real64>(min_i) * 60.0;

	/* Re-apply sign */
	deg_i = negative ? -deg_i : deg_i ; 
    }

    std::string asDms() const
    {
    /* Return a string of this angle in  ddDmmMss.sssS format */
	int deg=0 , min = 0;
        real64 sec = 0.0;
	bool negative = false;
        this->asDms(negative,deg , min , sec);
       
        int p = this->getprec();
        char c[255];

	// Handle deg == -0!
	if (negative)
	{
	    deg = -deg;
	    sprintf( c , "-%02d&nbsp;%02d&nbsp;%0*.*f" , deg , min , p+2 , p , sec );
	}
	else
	{
	    sprintf( c , "%02d&nbsp;%02d&nbsp;%0*.*f" , deg , min , p+2 , p , sec );
	}
        return std::string(c);    
    }

    std::string asAbrevDms() const
    {
    /* Return a string of this angle in  ddDmmMss.sssS format */
	int deg=0 , min = 0;
        real64 sec = 0.0;
	bool negative = false;
        this->asDms(negative,deg , min , sec);
       
        int p = this->getprec();
        char c[255];
	if (negative)
	{
	    if (deg)
	    {
		deg = -deg;
		sprintf( c , "-%02d&nbsp;%02d&nbsp;%0*.*f" , deg , min , p+2 , p , sec );
	    }
	    else if (min)
	    {
		sprintf( c , "-%02d&nbsp;%0*.*f" , min , p+2 , p , sec );
	    }
	    else 
	    {
		sprintf( c , "-%*.*f" , p+2 , p , sec );
	    }
	}
	else
	{
	    if (deg)
		sprintf( c , "%02d&nbsp;%02d&nbsp;%0*.*f" , deg , min , p+2 , p , sec );
	    else if (min)
		sprintf( c , "%02d&nbsp;%0*.*f" , min , p+2 , p , sec );
	    else 
		sprintf( c , "%*.*f" , p+2 , p , sec );
	}

        return std::string(c);    
    }


    real64 asDecHours() const
    {
	return this->get() / 15.0;
    }

    /* operator<< control flags
     * These are static because I most often use them ahead of a loop
     * eg   x.ashms(true); loop( cout << *(angle_iterator)...)
     */

    /* Output angles as hhHmmMss.sssS */
    Degrees::Format format() const { return _format; }
    Degrees::Format format(const Degrees::Format s) { Degrees::Format t = _format ; _format = s ; return t; }
    bool ashms() const { return _format == HMS; }

    /* Output angles as ddDmmMss.sssS */
    bool asdms() const { return _format == DMS; } 

    /* Output angles as decimal degrees */
    bool asdecdeg() const { return _format == DECIMALDEGREES; }

    /* Output angles as decimal hours */
    bool asdechours() const {  return _format == DECIMALHOURS; }

    friend std::ostream &operator<<(std::ostream& os , const Degrees& s)
    {
	std::streamsize oldsize = os.precision();
	int p = s.getprec();
	if(p)
	{
	    oldsize = os.precision(p);
	}

	if ( s.ashms() )
	{   
	    os << s.asHms();
	}
	else if ( s.asdms() )
	{
	    os << s.asDms();
	}
	else if ( s.asdechours() )
	{
	    real64 d = s.get() / 15.0;
	    os << d;
	}
	else
	{
	    os << s.get();
	}

	if(p)
	{
	    os.precision(oldsize);
	}
	return os;
    }
    private:
    static int _precision;
    static Degrees::Format _format;
};
/* statics */
//int Degrees::_precision = 0;
//Degrees::Format Degrees::_format = Degrees::DECIMALDEGREES;

class MakeDegrees
{
    public:
    Degrees operator()( const Degrees& s ) { return Degrees( s.get() ); }
    Degrees operator()( const real64&  s ) { return Degrees(s); }
    Degrees operator()( const char* s ) { return Degrees(s); }
    Degrees operator()( const std::string& s ) { return Degrees(s); }
};


#endif
