// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef PPSSUMM_EINDEX
#define PPSSUMM_EINDEX

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <iterator>
#include <sstream>
#include <cmath>

// SAS 1
#include "Dal.h"
#include "Tracer.h"
#include "errstr.h"

// PPSSUMM Type classes
#include "ppssumm_typedef.h"
#include "string_algo.h"
#include "key_algo.h"
#include "PpssummCommon.h"
#include "Filename_typedef.h"
#include "Filename.h"
#include "Instrument.h"
#include "Src_num.h"
#include "Band.h"
#include "Rgsorder.h"
#include "Degrees.h"
#include "Datamode.h"
#include "Pindex_typedef.h"
#include "Pindex.h"

/*
typedef std::string Mode;
typedef std::string Filter;
typedef int32 Exposure;
typedef bool8 Srcdet;
typedef bool8 Ssp;
typedef bool8 Flarescreened;
*/

class EpicmultiexposureproductRow
{
    public:
    EpicmultiexposureproductRow( const unsigned int v0 ,  const Name& v1 , const Content& v2 ) 
    : _row(v0) , _name(v1) , _content(v2) 
    {
    }

    bool operator<(const EpicmultiexposureproductRow& e) const
    {
	return _name < e.name();
    }

    friend ostream& operator<<( ostream& os , const EpicmultiexposureproductRow& r )
    {
	os  << "<tr>"
	    << "<td class=\"string\">" 
	    << r._name 
	    << "</td>"
	    << "<td class=\"string\">" 
	    << r._content 
	    << "</td>"
	    << "<td class=\"string\">" 
	    << r._name.band().energy_limits() 
	    << "</td>"
	    << "</tr>"
	    << "\n"
	;
	return os;
    }
    Name name() const { return _name; }
    Content content() const { return _content; }

    private:
    unsigned int _row;
    Name _name;
    Content _content;
};

class Epicmultiexposureproduct : public std::vector<EpicmultiexposureproductRow>
{
    public:
    Epicmultiexposureproduct() : std::vector<EpicmultiexposureproductRow>()  {}
    Epicmultiexposureproduct(const Pindex& pindex) : std::vector<EpicmultiexposureproductRow>() 
    {
    //errstr << "Epicmultiexposureproduct: begin" << Sas::message(Msg::App, Msg::Sparse);
	const Epicobs* epicobs = pindex.epicobs();
	unsigned long row=0;
	for ( Epicobs::const_iterator CI = epicobs->begin()
	    ; CI != epicobs->end()
	    ; ++CI
	){
	    this->push_back( EpicmultiexposureproductRow( row++ , CI->name() , CI->content() ) );
	}

    //errstr << "Epicmultiexposureproduct: sort" << Sas::message(Msg::App, Msg::Sparse);
	
    //if (this->begin() == this->end()) //errstr << "Epicmultiexposureproduct: sort - nothing to sort." << Sas::message(Msg::App, Msg::Sparse);
    //else
    std::sort(this->begin(), this->end() , std::less<EpicmultiexposureproductRow>() );
    //errstr << "Epicmultiexposureproduct: end" << Sas::message(Msg::App, Msg::Sparse);
    }

    void title( ostream& os ) const
    {
	os  << "<tr><th class=\"string\">Filename</th><th class=\"string\">Content</th><th =class\"string\">Band</th></tr>"
	    << "\n"
	;
    }
    friend ostream& operator<<( ostream& os , const Epicmultiexposureproduct& r )
    {
	r.title(os);
	for( Epicmultiexposureproduct::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }
    private:
};

class EpicexposuresprocessedRow
{
    public:
    EpicexposuresprocessedRow(
	const unsigned int v0
	,const Instrument& v1
	,const Exp_id& v2
	,const Scheduled& v3
	,const Mode& v4
	,const Datamode& v5
	,const Ep_filter& v6
	,const Position& v7
	,const Exp_duration& v8
	,const Exposure& v9
	,const Srcdet& v12
	,const Ssp& v13
	,const Flarescreened& v14
	,const Exp_start& v10
	,const Exp_stop& v11
    ) : _row(v0)
	, _instrument(v1)
	, _exp_id(v2)
	, _scheduled(v3)
	, _mode(v4)
	, _datamode(v5)
	, _ep_filter(v6)
	, _position(v7)
	, _exp_duration(v8)
	, _exposure(v9)
	, _srcdet(v12)
	, _ssp(v13)
	, _flarescreened(v14)
	, _exp_start(v10)
	, _exp_stop(v11)
    {
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }

    bool8 select( const Instrument& i , const Exp_id& e ) const
    {
	return ( _instrument == i && _exp_id == e );
    }

    bool operator<( const EpicexposuresprocessedRow& r ) const
    {
	return (_instrument < r._instrument)
	    || ((_instrument == r._instrument) && (_exp_start < r._exp_start))
	;
    }
    
    friend ostream& operator<<( ostream& os , const EpicexposuresprocessedRow& r )
    {
	ppssumm_internal::toString<bool8> tostring;

	std::string href1 = ppssumm_internal::makekey( r._instrument.pagename() , r._instrument , r._exp_id );
	std::string href2 = r._instrument.pagename() + "#sourcesprocessed";
	std::ostream::fmtflags incflags = os.setf( std::ios::fixed );
	unsigned int prec = os.precision();
	os << "<tr>"
	    << "<td class=\"string\">" << r._instrument << "</td>"
	    << "<td class=\"string\">" <<  "<a href=\"" << href1 << "\">" << r._exp_id << "</a></td>"
	    << "<td class=\"string\">" << tostring(r._scheduled) << "</td>"
	    << "<td class=\"string\">" << r._mode << "</td>"
	    << "<td class=\"string\">" << r._datamode << "</td>"
	    << "<td class=\"string\">" << r._ep_filter << "</td>"
	    << "<td class=\"string\">" << r._position << "</td>"
	;
	os.precision(0);
	os  << "<td class=\"number\">" << r._exp_duration << "</td>"
	    << "<td class=\"number\">" << r._exposure << "</td>"
	    << "<td class=\"flag\">" << tostring(r._srcdet) << "</td>"
	    << "<td class=\"flag\"><a href=\"" << href2 << "\">" << tostring(r._ssp) << "</a></td>"
	    << "<td class=\"flag\">" << tostring(r._flarescreened) << "</td>"
	    << "<td class=\"date\">" << ppssumm_internal::truncTime(r._exp_start) << "</td>"
	    << "<td class=\"date\">" << ppssumm_internal::truncTime(r._exp_stop) << "</td>"
	    << "</tr>"
	    << "\n"
	;
	os.precision(prec);
	os.setf( incflags );
	return os;
    }

    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Scheduled _scheduled;
    Mode _mode;
    Datamode _datamode;
    Ep_filter _ep_filter;
    Position _position;
    Exp_duration _exp_duration;
    Exposure _exposure;
    Srcdet _srcdet;
    Ssp _ssp;
    Flarescreened _flarescreened;
    Exp_start _exp_start;
    Exp_stop _exp_stop;
};

class Epicexposuresprocessed : public std::vector<EpicexposuresprocessedRow>
{
public:
    Epicexposuresprocessed() : std::vector<EpicexposuresprocessedRow>() {}
    Epicexposuresprocessed(const DataSet* fitsum , const Pindex& pindex ) : std::vector<EpicexposuresprocessedRow>() 
    {
	const Epicexp* epicexp = pindex.epicexp();
	int row=0;
	for ( Epicexp::const_iterator CI = epicexp->begin()
	    ; CI != epicexp->end()
	    ; ++CI
	){
	    std::string key = ppssumm_internal::makekey( CI->instrument() , CI->exp_id() );
	    sbmap_ci SI = _seen.find(key);
	    if ( SI == _seen.end() )
	    {
		_seen[key] = true;
		Exp_id exp_id = CI->exp_id();
		bool8 scheduled = false;
		if (! exp_id.empty() )
		{
		    scheduled = exp_id[0] == 'S';
		}
		
		std::string position("&nbsp");
		if ( CI->instrument() == "EMOS1" ||  CI->instrument() == "EMOS2"  )
		{
		    std::string tname = CI->instrument() == "EMOS1" ? "M1" : "M2";
		    stringlist  exp_id , filter_wheel , scheduled;
		    const Table* t = fitsum->table( tname );
		    stringColumn( t , "EXPID" , back_inserter( exp_id ));
		    stringColumn( t , "SCHEDULED" , back_inserter( scheduled ));
		    stringColumn( t , "FILTER_WHEEL" , back_inserter( filter_wheel ));

		    for(unsigned int i = 0,j = 0,k=0
			; i < exp_id.size() && j < filter_wheel.size() && k < scheduled.size() 
			; ++i , ++j , k++
		    ) {

	    
			std::stringstream ss;
			ss << (scheduled[k] == "Y" ? "S" : "U");
			ss << exp_id[i];
			if ( CI->exp_id() == ss.str() )
			{
			    position = filter_wheel[j];
			    break;
			}
		    }
		}
		this->push_back( EpicexposuresprocessedRow( row++
		    , CI->instrument() 
		    , CI->exp_id() 
		    , scheduled 
		    , CI->mode()
		    , CI->datamode()
		    , CI->ep_filter()
		    , position
		    , CI->exp_duration()
		    , CI->exposure()
		    , CI->srcdet()
		    , CI->ssp()
		    , CI->flarescreened()
		    , CI->exp_start()
		    , CI->exp_stop()
		));
	    }
	}

	std::sort(this->begin(), this->end() , std::less<EpicexposuresprocessedRow>() );
    }

    void title( ostream& os ) const
    {
	os << "<tr>"
	    << "<th class=\"string\">Inst.</th>"
	    << "<th class=\"string\">Exp.&nbsp;Id</th>"
	    << "<th class=\"string\">Sched</th>"
	    << "<th class=\"string\">Mode</th>"
	    << "<th class=\"string\">Datamode</th>"
	    << "<th class=\"string\">Filter</th>"
	    << "<th class=\"string\">Position</th>"
	    << "<th class=\"number\">Duration</th>"
	    << "<th class=\"number\">Exposure</th>"
	    << "<th class=\"number\">SrcDet</th>"
	    << "<th class=\"number\">SSP</th>"
	    << "<th class=\"number\">Flare Scrn</th>"
	    << "<th class=\"date\">Actual Start</th>"
	    << "<th class=\"date\">Actual End</th>"
	    << "</tr>"
	    << "\n"
	;
    }
    friend ostream& operator<<( ostream& os , const Epicexposuresprocessed& r )
    {
	r.title(os);
	for( Epicexposuresprocessed::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }

    const sbmap* seen() const { return &_seen; }

    private:
    sbmap _seen;
};

class EpicexposurespecificproductRow
{
    public:
    EpicexposurespecificproductRow( const unsigned int v0
	, const Instrument& v4
	, const Exp_id& v5
	,const Name& v1
	,const Content& v2
	,const Band& v3
    )
    : _row(v0)
    , _instrument(v4)
    , _exp_id(v5)
    , _name(v1)
    , _content(v2)
    , _band(v3)
    {

    }

    bool operator<( const EpicexposurespecificproductRow& r ) const
    {
	return ( _instrument < r._instrument )
	    || ( ( _instrument == r._instrument ) && ( _exp_id < r._exp_id ) )
	    || ( ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) && (_band < r._band) )
	    || ( ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) && (_band == r._band) && (_name < r._name))
	    || ( ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) && (_band == r._band) && (_name == r._name) && ( _content < r._content ))
	;
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }

    bool8 select( const Instrument& i , const Exp_id& e ) const
    {
	return ( _instrument == i && _exp_id == e );
    }

    friend ostream& operator<<( ostream& os , const EpicexposurespecificproductRow& r )
    {
	os  << "<tr><td class=\"string\">" 
	    << r._name 
	    << "</td>"
	    << "\n"
	    << "<td class=\"string\">" 
	    << r._content 
	    << "</td>"
	    << "\n"
	    << "<td class=\"string\">" 
	    << r._band.energy_limits() 
	    << "</td>"
	    << "\n"
	    << "</tr>"
	    << "\n"
	;
	return os;
    }

    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Name _name;
    Content _content;
    Band _band;
};

class Epicexposurespecificproduct : public std::vector<EpicexposurespecificproductRow>
{
    public:
    Epicexposurespecificproduct()
    : std::vector<EpicexposurespecificproductRow>()
    {
}

    Epicexposurespecificproduct( const Pindex& pindex )
    : std::vector<EpicexposurespecificproductRow>()
    {
	const Epicexp* ee = pindex.epicexp();
	unsigned int row = 0;
	for( Epicexp::const_iterator CI = ee->begin()
	    ; CI != ee->end()
	    ; ++CI
	){
	    this->push_back( EpicexposurespecificproductRow( row++
		, CI->instrument()
		, CI->exp_id()
		, CI->name()
		, CI->content()
		, CI->band()
	    ));
	}
	std::sort(this->begin(), this->end() , std::less<EpicexposurespecificproductRow>() );
    }

    void title( ostream& os ) const 
    {
	os  << "<tr><th class=\"string\">Filename</th><th class=\"string\">Content</th><th class=\"string\">Band</th></tr>"
	    << "\n"
	;
    }

    friend ostream& operator<<( ostream& os , const Epicexposurespecificproduct& r )
    {
	r.title(os);
	for( Epicexposurespecificproduct::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }
    private:

};

class EpicsourcesprocessedRow 
{
    public:
    EpicsourcesprocessedRow( unsigned int v0
	, const Instrument& _Instrument
	, const Exp_id& _Exp_id
	, const Src_num& _Src_num
	, const Ra_obj& _Ra_obj
	, const Dec_obj& _Dec_obj
	, const Radec_err& _Radec_err
	, const M1_cts& _M1_cts
	, const M1_cts_err& _M1_cts_err
	, const M2_cts& _M2_cts
	, const M2_cts_err& _M2_cts_err
	, const Pn_cts& _Pn_cts
	, const Pn_cts_err& _Pn_cts_err
	, const M1_tot_flux& _M1_tot_flux
	, const M1_tot_flux_err& _M1_tot_flux_err
	, const M2_tot_flux& _M2_tot_flux
	, const M2_tot_flux_err& _M2_tot_flux_err
	, const Pn_tot_flux& _Pn_tot_flux
	, const Pn_tot_flux_err& _Pn_tot_flux_err
	, const Ep_det_ml& _Ep_det_ml
	, const M1_det_ml& _M1_det_ml
	, const M2_det_ml& _M2_det_ml
	, const Pn_det_ml& _Pn_det_ml
	, const M1_ontime& _M1_ontime
	, const M2_ontime& _M2_ontime
	, const Pn_ontime& _Pn_ontime
	, const Ep_extent& _Ep_extent
	, const Ep_ext_err& _Ep_ext_err
	, const Ep_flag& _Ep_flag
	, const Srcdet& _Srcdet
    )
    : _row(_row)
	, _instrument(_Instrument)
	, _exp_id(_Exp_id)
	, _src_num(_Src_num)
	, _ra_obj(_Ra_obj)
	, _dec_obj(_Dec_obj)
	, _radec_err(_Radec_err)
	, _m1_cts(_M1_cts)
	, _m1_cts_err(_M1_cts_err)
	, _m2_cts(_M2_cts)
	, _m2_cts_err(_M2_cts_err)
	, _pn_cts(_Pn_cts)
	, _pn_cts_err(_Pn_cts_err)
	, _m1_tot_flux(_M1_tot_flux)
	, _m1_tot_flux_err(_M1_tot_flux_err)
	, _m2_tot_flux(_M2_tot_flux)
	, _m2_tot_flux_err(_M2_tot_flux_err)
	, _pn_tot_flux(_Pn_tot_flux)
	, _pn_tot_flux_err(_Pn_tot_flux_err)
	, _ep_det_ml(_Ep_det_ml)
	, _m1_det_ml(_M1_det_ml)
	, _m2_det_ml(_M2_det_ml)
	, _pn_det_ml(_Pn_det_ml)
	, _m1_ontime(_M1_ontime)
	, _m2_ontime(_M2_ontime)
	, _pn_ontime(_Pn_ontime)
	, _ep_extent(_Ep_extent)
	, _ep_ext_err(_Ep_ext_err)
	, _ep_flag(_Ep_flag)
	, _srcdet(_Srcdet)
    {
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }
    Src_num src_num() const { return _src_num; }

    bool8 select( const Instrument& i , const Exp_id& e , const Src_num& s ) const
    {
	return ( _instrument == i && _exp_id == e && _src_num == s );
    }

    bool operator<( const EpicsourcesprocessedRow& r ) const
    {
	// Sort order:  source number, instrument , source detection , exposure
	return  (_src_num < r._src_num)
	    || ( (_src_num == r._src_num) && (_instrument < r._instrument) ) 
	    || ( (_src_num == r._src_num) && (_instrument == r._instrument) && (_srcdet < r._srcdet) ) 
	    || ( (_src_num == r._src_num) && (_instrument == r._instrument) && (_srcdet == r._srcdet) && ( _exp_id < r._exp_id ) ) 
	;
    }

    std::string expanded_src_num() const
    {
	stringstream ss;
	ss << _src_num;
	ss << "&nbsp;(";
	_src_num.asHex(ss);
	ss << ")";

	return ss.str();
    }

    friend ostream& operator<<( ostream& os , const EpicsourcesprocessedRow r )
    {
	ppssumm_internal::toString<bool8> tostring;
	std::string href1 = ppssumm_internal::makekey( r._instrument.pagename()  , r._instrument , r._exp_id , r._src_num );
	std::string href2 = ppssumm_internal::makekey( r._instrument.pagename() , r._instrument , r._exp_id );
	std::ostream::fmtflags prev = os.setf( std::ios::fixed );
	unsigned int prec = os.precision();
	os  << "<tr>"
	    << "<td class=\"string\">" << "<a href=\"" << href1 << "\">" << r.expanded_src_num() << "</a></td>"
	    << "\n"
	    << "<td class=\"string\">" << r._instrument << "</td>"
	    << "\n"
	    << "<td class=\"string\">" << "<a href=\""<<  href2 << "\">" << r._exp_id << "</a></td>"
	    << "\n"
	;
	Degrees d(0.0);
	d.setprec(2);
	os  << "<td class=\"angle\">" << r._ra_obj.asHms() << "</td>"
	    << "\n"
	;

	d.setprec(1);
	os  << "<td class=\"angle\">" << r._dec_obj.asDms() << "</td>"
	    << "\n"
	    << "<td class=\"number\">" << r._radec_err.asAbrevDms() << "</td>"
	    << "\n"
	;
	  
	os.precision(1);
	if ( r._instrument == "EMOS1" )
	    os  << "<td class=\"number\">" << r._m1_cts << " &plusmn;" << r._m1_cts_err  << "</td>" << "\n";
	else if ( r._instrument == "EMOS2" )
	    os  << "<td class=\"number\">" << r._m2_cts << " &plusmn;" << r._m2_cts_err  << "</td>" << "\n";
	else if ( r._instrument == "EPN" )
	    os  << "<td class=\"number\">" << r._pn_cts << " &plusmn;" << r._pn_cts_err  << "</td>" << "\n";

	os.setf( ios::scientific , ios::basefield|ios::fixed );
	os.precision(3);
	if ( r._instrument == "EMOS1" )
	    os  << "<td class=\"number\">" << r._m1_tot_flux << " &plusmn;" << r._m1_tot_flux_err  << "</td>" << "\n";
	else if ( r._instrument == "EMOS2" )
	    os  << "<td class=\"number\">" << r._m2_tot_flux << " &plusmn;" << r._m2_tot_flux_err  << "</td>" << "\n";
	else if ( r._instrument == "EPN" )
	    os  << "<td class=\"number\">" << r._pn_tot_flux << " &plusmn;" << r._pn_tot_flux_err  << "</td>" << "\n";

	os.setf( std::ios::fixed );

	os.precision(0);
	os  << "<td class=\"number\">" << r._ep_det_ml << "</td>" << "\n";
	if ( r._instrument == "EMOS1" )
	    os  << "<td class=\"number\">" << r._m1_det_ml << "</td>" << "\n";
	else if ( r._instrument == "EMOS2" )
	    os  << "<td class=\"number\">" << r._m2_det_ml << "</td>" << "\n";
	else if ( r._instrument == "EPN" )
	    os  << "<td class=\"number\">" << r._pn_det_ml << "</td>" << "\n";

	os.precision(0);
	if ( r._instrument == "EMOS1" )
	    os  << "<td class=\"number\">" << r._m1_ontime << "</td>" << "\n";
	else if ( r._instrument == "EMOS2" )
	    os  << "<td class=\"number\">" << r._m2_ontime << "</td>" << "\n";
	else if ( r._instrument == "EPN" )
	    os  << "<td class=\"number\">" << r._pn_ontime << "</td>" << "\n";

	os.precision(0);
	os  << "<td class=\"number\">" << r._ep_extent;
	// If extent is 0 then ep_ext_err is NAN so don't write out
	// DJF 2009-06-11 Solaris C++ appears to have dropped std::isnan from cmath.
	// Rather than play games with includes fall back to the tried and trusted
	// comparison.  
	// NaN is the only value for which (NaN == NaN) will return false.
	//
	if ( r._ep_ext_err == r._ep_ext_err )
	    os <<  " &plusmn;" << r._ep_ext_err;
	os << "</td>" << "\n" ;

	os  << "<td class=\"flags\">" << r._ep_flag << "</td>" << "\n";

	os  << "<td class=\"flag\">" << tostring(r._srcdet) << "</td>"
	    << "\n"
	    << "</tr>"
	    << "\n"
	;
	os.setf(prev);
	os.precision(prec);
        return os;
    }
    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Src_num _src_num;
    Ra_obj _ra_obj;
    Dec_obj _dec_obj;
    Radec_err _radec_err;
    M1_cts _m1_cts;
    M1_cts_err _m1_cts_err;
    M2_cts _m2_cts;
    M2_cts_err _m2_cts_err;
    Pn_cts _pn_cts;
    Pn_cts_err _pn_cts_err;
    M1_tot_flux _m1_tot_flux;
    M1_tot_flux_err _m1_tot_flux_err;
    M2_tot_flux _m2_tot_flux;
    M2_tot_flux_err _m2_tot_flux_err;
    Pn_tot_flux _pn_tot_flux;
    Pn_tot_flux_err _pn_tot_flux_err;
    Ep_det_ml _ep_det_ml;
    M1_det_ml _m1_det_ml;
    M2_det_ml _m2_det_ml;
    Pn_det_ml _pn_det_ml;
    M1_ontime _m1_ontime;
    M2_ontime _m2_ontime;
    Pn_ontime _pn_ontime;
    Ep_extent _ep_extent;
    Ep_ext_err _ep_ext_err;
    Ep_flag _ep_flag;
    M1_flag _m1_flag;
    M2_flag _m2_flag;
    Pn_flag _pn_flag;
    Srcdet _srcdet;
    
};

class Epicsourcesprocessed : public std::vector<EpicsourcesprocessedRow>
{
    public:
    Epicsourcesprocessed() : std::vector<EpicsourcesprocessedRow>() {}
    Epicsourcesprocessed( const Pindex& pindex ) : std::vector<EpicsourcesprocessedRow>()
    {
	sbmap seen;
	const Epicsrc* es = pindex.epicsrc();
	unsigned int row=0;
	for( Epicsrc::const_iterator CI = es->begin()
	    ; CI != es->end()
	    ; ++CI
	) {
	    std::string key = ppssumm_internal::makekey( CI->instrument() , CI->exp_id() , CI->src_num() );
	    if ( seen.find(key) == seen.end() )
	    {
		this->push_back ( EpicsourcesprocessedRow( row++
		    , CI->instrument()
		    , CI->exp_id()
		    , CI->src_num()
		    , CI->ra_obj()
		    , CI->dec_obj()
		    , CI->radec_err()
		    , CI->m1_cts() , CI->m1_cts_err()
		    , CI->m2_cts() , CI->m2_cts_err()
		    , CI->pn_cts() , CI->pn_cts_err()
		    , CI->m1_tot_flux() , CI->m1_tot_flux_err()
		    , CI->m2_tot_flux() , CI->m2_tot_flux_err()
		    , CI->pn_tot_flux() , CI->pn_tot_flux_err()
		    , CI->ep_det_ml()
		    , CI->m1_det_ml()
		    , CI->m2_det_ml()
		    , CI->pn_det_ml()
		    , CI->m1_ontime()
		    , CI->m2_ontime()
		    , CI->pn_ontime()
		    , CI->ep_extent()
		    , CI->ep_ext_err()
		    , CI->ep_flag()
		    , CI->srcdet()
		));
		seen[key] = true;
	    }
	}
	std::sort(this->begin(), this->end() , std::less<EpicsourcesprocessedRow>() );
	
    }
    void title( ostream& os ) const
    {
	os  << "<tr>"
	    << "<th class=\"string\">Source</th>"
	    << "<th class=\"string\">Inst.</th>"
	    << "<th class=\"string\">Exp.&nbsp;Id</th>"
	    << "<th class=\"angle\">RA</th>"
	    << "<th class=\"angle\">Dec</th>"
	    << "<th class=\"angle\">RADec Err</th>"
	    << "<th class=\"number\">Count</th>"
	    << "<th class=\"number\">Band&nbsp;8 Flux</th>"
	    << "<th class=\"number\">EPIC Det&nbsp;ML</th>"
	    << "<th class=\"number\">Inst. Det&nbsp;ML</th>"
	    << "<th class=\"number\">Ontime</th>"
	    << "<th class=\"number\">EPIC Extent</th>"
	    << "<th class=\"flags\">EPIC Flags</th>"
	    << "<th class=\"flag\">Srcdet</th>"
	    << "</tr>"
	    << "\n"
	;
	os  << "<tr>"
	    << "<th class=\"unit\">Dec (Hex)</th>"
	    << "<th class=\"unit\">&nbsp</th>"
	    << "<th class=\"unit\">&nbsp</th>"
	    << "<th class=\"unit\">[hms]</th>"
	    << "<th class=\"unit\">[&deg; &prime; &Prime;]</th>"
	    << "<th class=\"unit\">[&Prime;]</th>"
	    << "<th class=\"unit\">&nbsp;</th>"
	    << "<th class=\"unit\">[erg/s/cm<sup>2</sup>]</th>"
	    << "<th class=\"unit\">&nbsp;</th>"
	    << "<th class=\"unit\">&nbsp;</th>"
	    << "<th class=\"unit\">[s]</th>"
	    << "<th class=\"unit\">[image pixels]</th>"
	    << "<th class=\"unit\">&nbsp</th>"
	    << "<th class=\"unit\">&nbsp</th>"
	    << "</tr>"
	    << "\n"
	;

    }

    friend ostream& operator<<( ostream& os , const Epicsourcesprocessed r )
    {
	r.title(os);
	for( Epicsourcesprocessed::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }
};

class EpicsourcespecificproductRow
{
    public:
    EpicsourcespecificproductRow( const unsigned int v0
	, const Instrument& v4
	, const Exp_id& v5
	, const Src_num& v6
	,const Name& v1
	,const Content& v2
	,const Band& v3
    )
    : _row(v0)
    , _instrument(v4)
    , _exp_id(v5)
    , _src_num(v6)
    , _name(v1)
    , _content(v2)
    , _band(v3)
    {
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }
    Src_num src_num() const { return _src_num; }

    bool8 select( const Instrument& i , const Exp_id& e , const Src_num& s ) const
    {
	return ( _instrument == i && _exp_id == e && _src_num == s );
    }
 
    bool operator<( const EpicsourcespecificproductRow& r ) const
    {
	return  ( _src_num < r._src_num )
	    || ( ( _src_num == r._src_num ) &&  (_name < r._name) )
	    || ( ( _src_num == r._src_num ) &&  (_name == r._name) && ( _instrument < r._instrument ) )
	    || ( ( _src_num == r._src_num ) &&  (_name == r._name) && ( _instrument == r._instrument ) && ( _exp_id < r._exp_id ) )
	    || ( ( _src_num == r._src_num ) &&  (_name == r._name) && ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) &&  (_band < r._band))
	    || ( ( _src_num == r._src_num ) &&  (_name == r._name) && ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) 
		&&  (_band == r._band) && ( _content < r._content ))
	;
    }

    friend ostream& operator<<( ostream& os , const EpicsourcespecificproductRow& r )
    {
	os  << "<tr>"
	    << "<td class=\"string\">" 
	    << r._name 
	    << "</td>"
	    << "<td class=\"string\">" 
	    << r._content 
	    << "</td>"
	    << "<td class=\"string\">" 
	    << r._band.energy_limits() 
	    << "</td>"
	    << "</tr>"
	    << "\n"
	;
	return os;
    }

    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Src_num _src_num;
    Name _name;
    Content _content;
    Band _band;
};

class Epicsourcespecificproduct : public std::vector<EpicsourcespecificproductRow>
{
    public:
    Epicsourcespecificproduct() : std::vector<EpicsourcespecificproductRow>() {}
    Epicsourcespecificproduct( const Pindex& pindex ) : std::vector<EpicsourcespecificproductRow>()
    {
	const Epicsrc* es = pindex.epicsrc();
	unsigned int row = 0;
	for( Epicsrc::const_iterator CI = es->begin()
	    ; CI != es->end()
	    ; ++CI
	){
	    this->push_back( EpicsourcespecificproductRow( row++
		, CI->instrument()
		, CI->exp_id()
		, CI->src_num()
		, CI->name()
		, CI->content()
		, CI->band()
	    ));
	}
	
	std::sort(this->begin(), this->end() , std::less<EpicsourcespecificproductRow>() );
    }

    void title( ostream& os ) const
    {
	os  << "<tr><th class=\"string\">Filename</th><th class=\"string\">Content</th><th class=\"string\">Band</th></tr>"
	    << "\n"
	;
    }
    friend ostream& operator<<( ostream& os , const Epicsourcespecificproduct r )
    {
	r.title(os);
	for( Epicsourcespecificproduct::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    
	    os << *CI;
	}
	return os;
    }

    private:

};

#endif
