// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include "ObsSummary.h"

ObsSummary::ObsSummary(std::string infile , DataSet* ds) 
{

    ifs_.open(infile.c_str());

    if (!ifs_) errstr << "Bad open on infile" << Sas::fatal("Sum2Fit:BADINFILE");
    ods_ = ds;


    //preread();

    std::string s = readline();
    while (s != "EOF" && s != "ERR") 
    {
        if (s == "OBSERVATION" || s == "SLEW") OBSERVATION();
        else if (s == "FILES") FILES();
        else if (s == "CONFIGURATION") CONFIGURATION();
        else if (s == "INSTRUMENT") INSTRUMENT();
        else if (s == "PROPOSAL") PROPOSAL();
        else if (s == "DATA QUALITY") DATA_QUALITY();

        s = readline();
    }

    postread();
    writeAttributes();

    ifs_.close();

    return;
}

void ObsSummary::preread()
{
    /* The comments at the top of the summary file contain useful information
     * which will be added to the keywords of the dataset
     */
    std::string line;
    SAttribute odfingver("ODFINGVR" , "NA"  , "odfingest version" );
    SAttribute odfingdt("ODFINGDT" , "NA"  , "odfingest date" );
    SAttribute sasmanifest("SASMAN" , "NA"  , "sas manifest containing odfingest" );
    SAttribute fix_pn_mode("FIXPNMOD" ,  "N" , "True if fix_pn_mode applied" );

    
    getline(ifs_,line);

    while ( ifs_ && line.find("OBSERVATION") == std::string::npos)
    {

        if (line.find("odfingest") != std::string::npos) {

            std::string::size_type verF = line.find("(")+1;
            std::string::size_type verL = line.find(")") - verF ;
            odfingver.value( stripws ( line.substr( verF , verL ) ) );

            std::string::size_type sasF = line.find("[")+1;
            std::string::size_type sasL = line.find("]") - sasF;
            sasmanifest.value( stripws( line.substr( sasF , sasL ) ) );

            std::string::size_type dtF = line.find("on") +2;
            std::string::size_type dtL = line.find_last_of("0123456789") - dtF;
            odfingdt.value( stripws( line.substr( dtF,dtL ) ) );
            
        }

        if (line.find("fix_pn_mode") != std::string::npos) {
            
            fix_pn_mode.value("Y");
        }

        getline(ifs_,line);

    }
    rtn_string = "OBSERVATION";

    addAtt(0 , odfingver);
    addAtt(0 , odfingdt);
    addAtt(0 , sasmanifest);
    addAtt(0 , fix_pn_mode);
    
    return;
}

void ObsSummary::postread() { return; }

std::string ObsSummary::readline(bool emptylinessignificant) 
{

    // Two comment lines with significant information are handled by the function preread.
    // Whole line comments are ignored.
    // Parameter comments can be used to suppliment keywords so we keep them.
    // empty lines can be siginificant (such as in the PROPOSAL section).
    // leading/trailing whitespace is removed from the name, value and comment.

    std::string line;
    if (rtn_string.length()) {
        line = rtn_string;
        rtn_string="";
        return line;
    }
    
    if (ifs_) getline(ifs_,line) ;

    while (ifs_) {
	errstr << line << Sas::message(Msg::App, Msg::Sparse);	 
        if ( (line.find("/") == std::string::npos)
            || line.find_first_not_of("/") < line.find("/")        
        ) {
            line = stripws(line);
            if ( ! (emptylinessignificant || line.length())  ) break;
            
            return line;
        }
        getline(ifs_,line);

    }

    return std::string("EOF");
}


void ObsSummary::writeAttributes() 
{
    // Write the collected attributes to the data set
    
    for (unsigned int i = 0 ; i < ods_->blocks() ; i++) {

        Table* tbl = ods_->table(i);
        for (multimap< unsigned int , SAttribute >::const_iterator CI = Atts_.lower_bound(i) ; CI != Atts_.upper_bound(i) ; CI++ )
            addto(tbl, CI->second); 
    }
    return;
}    

void ObsSummary::OBSERVATION() 
{
// OBSERVATIONsection
// Assume readline will return line after OBSERVATION.
    
    errstr << "OBSERVATION section" << Sas::message(Msg::App, Msg::Sparse);

    STable obs("OBSERVAT");
    obs.push_back( SColumn( MakeAttribute( readline() , "OBS_ID") ) );
    obs.push_back( SColumn( MakeAttribute( readline() , "REVOLUTION") ) );
    obs.push_back( SColumn( MakeAttribute( readline() , "OBS_START") ) );
    obs.push_back( SColumn( MakeAttribute( readline() , "OBS_STOP") ) );

    addto(ods_,obs);
    errstr << "OBSERVATION section done." << Sas::message(Msg::App, Msg::Sparse);

}

void ObsSummary::FILES() 
{
// FILES section
// Assume readline will return line after FILES.
    

/* In older format ODF summary files an INSTRUMENT section follows the
 * FILES section rather than the CONFIGURATION section.
    
 * Width of character column, CD names are max 32 chars including trailing \0
 * A FITS column will record  31 character but not the trailing \0.

 */
    errstr << "FILES section" << Sas::message(Msg::App, Msg::Sparse);

    STable files("FILES");


    SAttribute f1 = MakeAttribute(readline(),"FILES");
    long rows = atol ( f1.value().c_str() ) ;
    
    while ( rows-- ) {
        
        files.push_back( MakeAttribute(readline(),"FILENAME") ); 
    }

    
    addto(ods_,files);

    Table* tbl = ods_->table("FILES");
    addto(tbl,f1);

    errstr << "FILES section done." << Sas::message(Msg::App, Msg::Sparse);
}

void ObsSummary::CONFIGURATION() 
{
    // CONFIGURATION section
    // Assume readline will return line after CONFIGURATION.

    errstr << "CONFIGURATION (CONFIG) section" << Sas::message(Msg::App, Msg::Sparse);
    STable config("CONFIG");

    SAttribute f1 = MakeAttribute(readline(),"CONFPRMS");
    long rows = atol ( f1.value().c_str() ) ;
    while ( rows-- ) 
        config.push_back(  MakeAttribute( readline() ) ); 

    addto(ods_,config);
    Table* tbl = ods_->table("CONFIG");
    addto(tbl,f1);
    
    errstr << "CONFIGURATION (CONFIG) section done." << Sas::message(Msg::App, Msg::Sparse);
}

void ObsSummary::INSTRUMENT() 
{
// an INSTRUMENT has EXPOSURE subsections
// an EXPOSURE has CONFIGURATION subsection

// Assume INSTRUMENT line has been read.
    errstr << "INSTRUMENT section" << Sas::message(Msg::App, Msg::Sparse);

    SAttribute inst = MakeAttribute(readline() , "INST");

    SAttribute active = MakeAttribute(readline() , "ACTIVE");
    SAttribute nrexp("NREXP", "0" , "Number of Exposures");

    // I don't trust the nrexp value.  Maybe I should  but I don't need to

    STable st( inst.value() );
    
    if (active.value() == "Y") {

        nrexp.value( MakeAttribute(readline() , "dummy").value() );

        std::string s = readline();
        while (s == "EXPOSURE" && s != "EOF") {
            STable exptbl("dummy");
            EXPOSURE(exptbl);
            st.insert(exptbl.begin() , exptbl.end());
            st.pad();

            s = readline();
        }
        rtn_string = s;

    }

    addto(ods_ , st); 

    Table* tbl = ods_->table(inst.value());
    addto( tbl , inst);
    addto( tbl , active);
    addto( tbl , nrexp);
    
    addAtt( 0 , SAttribute( inst.value() + active.name() , active.value() , "Is Instrument Active"  ));
    addAtt( 0 , SAttribute( inst.value() + nrexp.name() , nrexp.value() , "Expected number of exposures" ));
    errstr << "INSTRUMENT (" << inst.value() << ") section done." << Sas::message(Msg::App, Msg::Sparse);
    
}

void ObsSummary::EXPOSURE(STable& st) 
{
// an INSTRUMENT has EXPOSURE subsections
// an EXPOSURE has CONFIGURATION subsection

// Assume EXPOSURE line has been read.
    errstr << "EXPOSURE section" << Sas::message(Msg::App, Msg::Sparse);
    std::string names[] = {
        "EXPID"
        , "SCHEDULED"
        , "EXP_TYPE"
        , "COMMAND_EXPID"
        , "PROPOSED_EXPID"
        , "SCHEDULED_START"
        , "SCHEDULED_END"
        , "ACTUAL_START"
        , "ACTUAL_END"
        , ""
    };

    std::string* p = names;
    while ( p->length() ) {

        st.push_back(  MakeAttribute( readline() , *p++) );
    }
    
    std::string s = readline();

    if (s.find("CONFIGURATION") != std::string::npos) { 
        STable cnftbl("DUMMY");
        INSTCONFIG( cnftbl );
        st.insert(cnftbl.begin() , cnftbl.end());
    } else {
        rtn_string = s;
    }

    errstr << "EXPOSURE section done." << Sas::message(Msg::App, Msg::Sparse);


}

void ObsSummary::INSTCONFIG(STable& st) 
{
// an INSTRUMENT has EXPOSURE subsections
// an EXPOSURE has CONFIGURATION subsection

// Assume CONFIGURATION line has been read.

    errstr << "EXPOSURE CONFIGURATION section" << Sas::message(Msg::App, Msg::Sparse);
    SAttribute f1 = MakeAttribute( readline() , "PARMS" );
    unsigned long parms = static_cast<unsigned long>( atol( f1.value().c_str() )  ) ;
    
    Datamode dm;
    while ( parms-- ) 
    {
	SAttribute tmp( MakeAttribute( readline() ) );
        st.push_back( tmp );
	std::string tmp_s= tmp.name();
	if (tmp_s.find("DATA_MODE") != std::string::npos )
	{
	    dm.push_back( tmp.value() );
	}
    }
    stringstream ss;
    ss << dm;
    st.push_back( SAttribute("DATA_MODE" , ss.str() , "Sumamry data mode" ) );
    errstr << "EXPOSURE CONFIGURATION section done." << Sas::message(Msg::App, Msg::Sparse);
    
}

void ObsSummary::PROPOSAL() {

// PROPOSAL section
// Assume readline will return line after PROPOSAL.
    errstr << "PROPOSAL section" << Sas::message(Msg::App, Msg::Sparse);
    std::string names[] = {
        "GONAME"
        , "GOINSTITUTE"
        , "GOADDRLN1"
        , "GOADDRLN2"
        , "GOADDRLN3"
        , "GOADDRLN4"
        , "GOADDRLN5"
        , "GOADDRLN6"
        , "GOEMAIL"
        , "AO"
        , "SCIENCE_TYPE"
        , "TARGET"
        , "TARGET_RA"
        , "TARGET_DEC"
        , "PROPOSED_DURATION"
        , "ALT_NAMES"
        , "BORESIGHT_RA"
        , "BORESIGHT_DEC"
        , "POS_ANGLE_LC"
        , "POS_ANGLE_UC"
        , "POS_ANGLE_REF"
        , "M1_PRIORITY"
        , "M2_PRIORITY"
        , "PN_PRIORITY"
        , "R1_PRIORITY"
        , "R2_PRIORITY"
        , "OM_PRIORITY"
        , "" 
    };

    STable prop("PROPOSAL");
    std::string* p = names;
    while ( p->length() ) {

        prop.push_back(  MakeAttribute( readline() , *p++ ) );
    }
    addto(ods_,prop);
    
    errstr << "PROPOSAL section done." << Sas::message(Msg::App, Msg::Sparse);
}

void ObsSummary::DATA_QUALITY()
{

/* DATA QUALITY section
 * Turn all DATA_QUALITY parts into a single table.
 * Assume readline will return line after DATA QUALITY.
 */

    STable dquality("DQUALITY");

    const std::string names[] = {
        "INSTRUMENT"
        ,"DQ_EXPID"
        ,"DQ1"
        ,"DQ2"
        ,"DQ3"
        ,"DQ4"
        ,"DQ5"
        ,"DQ6"
        ,"SCHEDULED"
        ,""
    };

    const std::string comments[] = {
        "Instrument Id"
        ,"Exposure Id"
        ,"Number of packets that failed reception checking"
        ,"Number of event reports"
        ,"Number of PMS errors"
        ,"Number of Telemetry drops"
        ,"Total duration of Telemetry drops"
        ,"Number of Telemetry out of limits"
        ,"Scheduled flag"
        ,""
    };


    std::string s="DATA QUALITY";
    while (s == "DATA QUALITY" && s != "EOF") {
        const std::string* n = names;
        const std::string* c = comments;
        while ( n->length() ) {

            SAttribute f1 = MakeAttribute( readline(), *n++ );
            f1.comment( *c++ );
            if ( ! f1.value().length() ) f1.value( "NA" );
            dquality.push_back( f1 );    
        }
        dquality.pad();
        s = readline();
    } 

    addto(ods_,dquality);

}

