// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef PPSSUMM_OM_SUMMARY
#define PPSSUMM_OM_SUMMARY

#include <vector>
#include <string>
#include <iostream>
#include <algorithm>
#include <iterator>
#include "ppssumm_typedef.h"
#include "PpssummCommon.h"
#include "Pindex.h"
#include "Omindex.h"
#include "ppssumm_algo.h"
#include "string_algo.h"
#include "key_algo.h"
#include "STime.h"
#include "Dal.h"
#include "param.h"
#include "Tracer.h"
#include "errstr.h"

class OmSummary 
{
    public:
    OmSummary( Pindex& pindex ) :
	_ommultiexposureproduct(pindex) 
	,_omexposuresprocessed(pindex)
	,_omexposurespecificproduct(pindex)
	,_omsourcesprocessed(pindex)
	,_omsourcespecificproduct(pindex)
    { }

    friend std::ostream &operator<<(std::ostream& os , const OmSummary& es) {

	// Page menu
	os  << "<div id=\"pagemenu\">\n"
	    << "<ul>\n"
	    << "<li><a href=\"#mulitexposureproduct\">Multi-Exposure Products</a></li>\n"
	    << "<li><a href=\"#exposuresprocessed\">Exposures processed by PPS</a></li>\n"
	    << "<li><a href=\"#sourcesprocessed\">Sources processed by PPS</a></li>\n"
	    << "</ul>\n"
	    << "<br clear=\"left\" />"
	    << "<hr />\n"
	    << "</div>"
	;
	// Now the fun begins

	os  << "<div id=\"normaltable\">"  << "\n"
	    << "<a name=\"mulitexposureproduct\">&nbsp;</a>"
	    << "<h3>Multi-Exposure Products</h3>"
	;
	if (es._ommultiexposureproduct.size())
	{
	    os	<< "<table class=\"ommultiexposureproduct_table\">"
		<< es._ommultiexposureproduct
		<< "</table>"
	    ;
	}
	else
	{
	    os << "<p class=\"noprod\">No multi-exposure products were produced.</p>";
	}
	os << "<hr />";
	os << "</div>";

	// EXPOSURES

	os << "<div id=\"exposuresprocessed\">" << "\n";
	os << "<a name=\"exposuresprocessed\">&nbsp;</a>" << "\n";
	os << "<h3>Exposures processed by PPS</h3>" << "\n";
	if (es._omexposuresprocessed.size())
	{
	    // Exposures processed table
	    os << "<div id=\"widetable\">" << "\n";
	    os << "<table class=\"omexposureprocessed_table\">" << "\n";
	    os << es._omexposuresprocessed;
	    os << "</table>" << "\n";
	    os << "<hr />" << "\n";
	    os << "</div>" << "\n";

	    // Exposure specific products
	    os << "<div id=\"exposurespecificproducts\">" << "\n";
	    os << "<h3>Exposure Specific Products</h3>" << "\n";
	    if( es._omexposurespecificproduct.size() )
	    {
		for ( Omexposuresprocessed::const_iterator CI = es._omexposuresprocessed.begin()
		    ; CI != es._omexposuresprocessed.end()
		    ; ++CI
		){

		    // Select exposurespecificproduct for this instrument and exposure.
		    Omexposurespecificproduct tmp;
		    for( Omexposurespecificproduct::const_iterator CI2 = es._omexposurespecificproduct.begin()
			; CI2 != es._omexposurespecificproduct.end()
			; ++CI2
		    ){
			if( CI2->select( CI->instrument() , CI->exp_id() ) ) tmp.push_back( *CI2 );
		    }

		    // Repeat exposure row ahead of exposure specific product table
		    os << "<div id=\"widetable\">" << "\n";
		    os << "<a name=\"" << ppssumm_internal::makekey( CI->instrument() , CI->exp_id() ) << "\">&nbsp;</a>" << "\n";
		    os << "<table class=\"omexposurespecific_toptable\">" << "\n";
		    es._omexposuresprocessed.title(os);
		    os << *CI;
		    os << "</table>" << "\n";
		    os << "</div>" << "\n";

		    // Write out exposurespecificproduct table
		    if (tmp.size())
		    {
			std::sort(tmp.begin(), tmp.end() , std::less<OmexposurespecificproductRow>() );
		    os << "<div id=\"normaltable\">" << "\n";
			os << "<table class=\"omexposurespecific_toptable\">" << "\n";
			os << tmp;
			os << "</table>" << "\n";
		    os << "</div>" << "\n";
		    }
		    else
		    {
			os << "<p class=\"noprod\">No exposure specific products were produced.</p>" << "\n";
		    }
		}
	    }
	    else
	    {
		os << "<p class=\"noprod\">No exposure specific products were produced.</p>" << "\n";
	    }
	    os << "<hr />" << "\n";
	    os << "</div>" << "\n";
	}
	else
	{
	    os << "<p class=\"noprod\">No OM exposures processed by the PPS.</p>" << "\n";
	    os << "<hr />" << "\n";
	}
	os << "</div>" << "\n";


	// SOURCES

	os << "<div class=\"omsourcesprocessed\">" << "\n";
	os << "<a name=\"sourcesprocessed\">&nbsp;</a>" << "\n";
	os << "<h3>Sources processed by PPS</h3>" << "\n";
	if (es._omsourcesprocessed.size())
	{
	    // Sources processed table
	    os << "<div id=\"widetable\">" << "\n";
	    os << "<table class=\"omsourceprocessed_table\">" << "\n";
	    os << es._omsourcesprocessed;
	    os << "</table>" << "\n";
	    os << "<hr />" << "\n";
	    os << "</div>" << "\n";

	    // Source specific products
	    os << "<div class=\"omsourcespecificproduct\">" << "\n";
	    os << "<h3>Source Specific Products</h3>" << "\n";
	    if( es._omsourcespecificproduct.size() )
	    {	
		for ( Omsourcesprocessed::const_iterator CI = es._omsourcesprocessed.begin()
		    ; CI != es._omsourcesprocessed.end()
		    ; ++CI
		){
		    // Select source specific product for this instrument and source.
		    Omsourcespecificproduct tmp;
		    for( Omsourcespecificproduct::const_iterator CI2 = es._omsourcespecificproduct.begin()
			; CI2 != es._omsourcespecificproduct.end()
			; ++CI2
		    ){
			if( CI2->select( CI->instrument() , CI->exp_id(), CI->src_num() ) ) tmp.push_back( *CI2 );
		    }

		    // Repeat source row ahead of source specific product table
		    os  << "<div id=\"widetable\">" << "\n"
			<< "<a name=\"" <<  ppssumm_internal::makekey( CI->instrument() , CI->exp_id() , CI->src_num() ) << "\">&nbsp;</a>" << "\n"
			<< "<table class=\"omsourcespecific_toptable\">" << "\n"
		    ;
		    es._omsourcesprocessed.title(os);
		    os	<< *CI
			<< "</table>" << "\n"
			<< "</div>" << "\n"
		    ;

		    // Write out sourcespecificproduct table
		    if (tmp.size())
		    {
			std::sort(tmp.begin(), tmp.end() , std::less<OmsourcespecificproductRow>() );
			os  << "<div id=\"normaltable\">" << "\n"
			    << "<table class=\"omsourcespecific_toptable\">" << "\n"
			    << tmp
			    << "</table>" << "\n"
			    << "</div>"
			;
		    }
		    else
		    {
			os << "<p class=\"noprod\">No source specific products were produced.</p>" << "\n"
			;
		    }
		}
	    }
	    else
	    {
		os  << "<p class=\"noprod\">No source specific products were produced.</p>" << "\n"
		;
	    }
	    os  << "<hr />" << "\n"
		<< "</div>" << "\n"
	    ;
	}
	else
	{
	    os  << "<p class=\"noprod\">No OM sources processed by the PPS.</p>" << "\n"
	    ;
	    os  << "<hr />" << "\n";
	}
	os  << "</div>" << "\n";

        return os;
    }
   
    const Omexposuresprocessed* omexposuresprocessed() const { return &_omexposuresprocessed; }
 
    private:
    Ommultiexposureproduct _ommultiexposureproduct;
    Omexposuresprocessed _omexposuresprocessed;
    Omexposurespecificproduct _omexposurespecificproduct;
    Omsourcesprocessed _omsourcesprocessed;
    Omsourcespecificproduct _omsourcespecificproduct;

};

#endif
