// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef PPSSUMM_OMINDEX
#define PPSSUMM_OMINDEX

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <iterator>
#include <sstream>
#include <cmath>

// SAS 1
#include "Dal.h"
#include "Tracer.h"
#include "errstr.h"

// PPSSUMM Type classes
#include "ppssumm_typedef.h"
#include "string_algo.h"
#include "key_algo.h"
#include "PpssummCommon.h"
#include "Filename_typedef.h"
#include "Filename.h"
#include "Instrument.h"
#include "Src_num.h"
#include "Band.h"
#include "Rgsorder.h"
#include "Degrees.h"
#include "Datamode.h"
#include "Pindex_typedef.h"
#include "Pindex.h"

class OmmultiexposureproductRow
{
    public:
    OmmultiexposureproductRow( const unsigned int v0 ,  const Name& v1 , const Content& v2 ) 
    : _row(v0) , _name(v1) , _content(v2) 
    {
    }

    bool operator<(const OmmultiexposureproductRow& e) const
    {
	return _name < e.name();
    }

    friend ostream& operator<<( ostream& os , const OmmultiexposureproductRow& r )
    {
	os  << "<tr>"
	    << "<td class=\"string\">" 
	    << r._name 
	    << "</td>"
	    << "<td class=\"string\">" 
	    << r._content 
	    << "</td>"
	    << "</tr>"
	    << "\n"
	;
	return os;
    }
    Name name() const { return _name; }
    Content content() const { return _content; }

    private:
    unsigned int _row;
    Name _name;
    Content _content;
};

class Ommultiexposureproduct : public std::vector<OmmultiexposureproductRow>
{
    public:
    Ommultiexposureproduct() : std::vector<OmmultiexposureproductRow>()  {}
    Ommultiexposureproduct(const Pindex& pindex) : std::vector<OmmultiexposureproductRow>() 
    {
	const Omobs* omobs = pindex.omobs();
	unsigned long row=0;
	for ( Omobs::const_iterator CI = omobs->begin()
	    ; CI != omobs->end()
	    ; ++CI
	){
	    this->push_back( OmmultiexposureproductRow( row++ , CI->name() , CI->content() ) );
	}

	std::sort(this->begin(), this->end() , std::less<OmmultiexposureproductRow>() );
    }

    void title( ostream& os ) const
    {
	os  << "<tr><th class=\"string\">Filename</th><th class=\"string\">Content</th></tr>"
	    << "\n"
	;
    }
    friend ostream& operator<<( ostream& os , const Ommultiexposureproduct& r )
    {
	r.title(os);
	for( Ommultiexposureproduct::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }
    private:
};

class OmexposuresprocessedRow
{
    public:
    OmexposuresprocessedRow(
	const unsigned int v0
	,const Instrument& v1
	,const Exp_id& v2
	,const Scheduled& v3
	,const Detector& v15
	,const Mode& v4
	,const Datamode& v5
	,const Om_filter& v6
	,const Exp_duration& v8
	,const Exp_start& v10
	,const Exp_stop& v11
    ) : _row(v0)
	, _instrument(v1)
	, _exp_id(v2)
	, _scheduled(v3)
	, _detector(v15)
	, _mode(v4)
	, _datamode(v5)
	, _om_filter(v6)
	, _exp_duration(v8)
	, _exp_start(v10)
	, _exp_stop(v11)
    {
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }

    bool8 select( const Instrument& i , const Exp_id& e ) const
    {
	return ( _instrument == i && _exp_id == e );
    }

    bool operator<( const OmexposuresprocessedRow& r ) const
    {
	return (_exp_start < r._exp_start);
    }
    
    friend ostream& operator<<( ostream& os , const OmexposuresprocessedRow& r )
    {
	ppssumm_internal::toString<bool8> tostring;

	std::string href = ppssumm_internal::makekey( r._instrument.pagename() , r._instrument , r._exp_id );
	std::ostream::fmtflags incflags = os.setf( std::ios::fixed );
	unsigned int prec = os.precision();
	os << "<tr>"
	    << "<td class=\"string\">" << r._instrument << "</td>"
	    << "<td class=\"string\">" <<  "<a href=\"" << href << "\">" << r._exp_id << "</a></td>"
	    << "<td class=\"string\">" << tostring(r._scheduled) << "</td>"
	    << "<td class=\"string\">" << r._detector << "</td>"
	    << "<td class=\"string\">" << r._mode << "</td>"
	    << "<td class=\"string\">" << r._datamode << "</td>"
	    << "<td class=\"string\">" << r._om_filter << "</td>"
	;
	os.precision(0);
	os  << "<td class=\"number\">" << r._exp_duration << "</td>"
	    << "<td class=\"date\">" << ppssumm_internal::truncTime(r._exp_start) << "</td>"
	    << "<td class=\"date\">" << ppssumm_internal::truncTime(r._exp_stop) << "</td>"
	    << "</tr>"
	    << "\n"
	;
	os.precision(prec);
	os.setf( incflags );
	return os;
    }

    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Scheduled _scheduled;
    Detector _detector;
    Mode _mode;
    Datamode _datamode;
    Om_filter _om_filter;
    Exp_duration _exp_duration;
    Exp_start _exp_start;
    Exp_stop _exp_stop;
};

class Omexposuresprocessed : public std::vector<OmexposuresprocessedRow>
{
public:
    Omexposuresprocessed() : std::vector<OmexposuresprocessedRow>() {}
    Omexposuresprocessed( const Pindex& pindex ) : std::vector<OmexposuresprocessedRow>() 
    {

	int row=0;
	// EXPOSURES
	const Omexp* omexp = pindex.omexp();
	for ( Omexp::const_iterator CI = omexp->begin()
	    ; CI != omexp->end()
	    ; ++CI
	){
	    std::string key = ppssumm_internal::makekey( CI->instrument() , CI->exp_id() );
	    sbmap_ci SI = _seen.find(key);
	    if ( SI == _seen.end() )
	    {
		_seen[key] = true;
		Exp_id exp_id = CI->exp_id();
		bool8 scheduled = false;
		if (! exp_id.empty() )
		{
		    scheduled = exp_id[0] == 'S';
		}
		
		this->push_back( OmexposuresprocessedRow( row++
		    , CI->instrument() 
		    , CI->exp_id() 
		    , scheduled 
		    , CI->detector()
		    , CI->mode()
		    , CI->datamode()
		    , CI->om_filter()
		    , CI->exp_duration()
		    , CI->exp_start()
		    , CI->exp_stop()
		));
	    }
	}
	// SCIENCE WINDOWS
	// It is possible for an exposure to generate only SW product so we need to 
	// process the OMSW table too but we do not want duplicates rows for each SW.
	const Omsw* omsw = pindex.omsw();
	for ( Omsw::const_iterator CI = omsw->begin()
	    ; CI != omsw->end()
	    ; ++CI
	){
	    std::string key = ppssumm_internal::makekey( CI->instrument() , CI->exp_id() );
	    sbmap_ci SI = _seen.find(key);
	    if ( SI == _seen.end() )
	    {
		_seen[key] = true;
		Exp_id exp_id = CI->exp_id();
		bool8 scheduled = false;
		if (! exp_id.empty() )
		{
		    scheduled = exp_id[0] == 'S';
		}

		this->push_back( OmexposuresprocessedRow( row++
		    , CI->instrument() 
		    , CI->exp_id() 
		    , scheduled 
		    , CI->detector()
		    , CI->mode()
		    , CI->datamode()
		    , CI->om_filter()
		    , CI->exp_duration()
		    , CI->exp_start()
		    , CI->exp_stop()
		));

	    }
	}

	std::sort(this->begin(), this->end() , std::less<OmexposuresprocessedRow>() );
    }

    void title( ostream& os ) const
    {
	os << "<tr>"
	    << "<th class=\"string\">Inst</th>"
	    << "<th class=\"string\">ExpID</th>"
	    << "<th class=\"string\">Sched</th>"
	    << "<th class=\"string\">Detector</th>"
	    << "<th class=\"string\">Mode</th>"
	    << "<th class=\"string\">Data Mode</th>"
	    << "<th class=\"string\">Filter</th>"
	    << "<th class=\"number\">Duration</th>"
	    << "<th class=\"date\">Actual Start</th>"
	    << "<th class=\"date\">Actual End</th>"
	    << "</tr>"
	    << "\n"
	;
    }
    friend ostream& operator<<( ostream& os , const Omexposuresprocessed& r )
    {
	r.title(os);
	for( Omexposuresprocessed::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }

    const sbmap* seen() const { return &_seen; }
    private:
    sbmap _seen;
};

class OmexposurespecificproductRow
{
    public:
    OmexposurespecificproductRow( const unsigned int v0
	, const Instrument& v4
	, const Exp_id& v5
	,const Name& v1
	,const Content& v2
	,const Osw_id& v3
    )
    : _row(v0)
    , _instrument(v4)
    , _exp_id(v5)
    , _name(v1)
    , _content(v2)
    , _osw_id(v3)
    {

    }

    bool operator<( const OmexposurespecificproductRow& r ) const
    {
	return ( _instrument < r._instrument )
	    || ( ( _instrument == r._instrument ) && ( _exp_id < r._exp_id ) )
	    || ( ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) && (_osw_id < r._osw_id) )
	    || ( ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) && (_osw_id == r._osw_id) && (_name < r._name))
	    || ( ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) && (_osw_id == r._osw_id) && (_name == r._name) && ( _content < r._content ))
	;
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }

    bool8 select( const Instrument& i , const Exp_id& e ) const
    {
	return ( _instrument == i && _exp_id == e );
    }

    friend ostream& operator<<( ostream& os , const OmexposurespecificproductRow& r )
    {
	os  << "<tr><td class=\"string\">" 
	    << r._name 
	    << "</td>"
	    << "\n"
	    << "<td class=\"string\">" 
	    << r._content 
	    << "</td>"
	    << "\n"
	    << "<td class=\"string\">" 
	    // We are only interested in numerical osw_id here.
	    << r._osw_id.osw_id()
	    << "</td>"
	    << "\n"
	    << "</tr>"
	    << "\n"
	;
	return os;
    }

    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Name _name;
    Content _content;
    Osw_id _osw_id;
};

class Omexposurespecificproduct : public std::vector<OmexposurespecificproductRow>
{
    public:
    Omexposurespecificproduct()
    : std::vector<OmexposurespecificproductRow>()
    {
    }

    Omexposurespecificproduct( const Pindex& pindex )
    : std::vector<OmexposurespecificproductRow>()
    {

	// Exposure level products have no OSWID so ecord an empty one.
	const Omexp* ee = pindex.omexp();
	unsigned int row = 0;
	for( Omexp::const_iterator CI = ee->begin()
	    ; CI != ee->end()
	    ; ++CI
	){
	    this->push_back( OmexposurespecificproductRow( row++
		, CI->instrument()
		, CI->exp_id()
		, CI->name()
		, CI->content()
		, -1
	    ));
	}
	std::sort(this->begin(), this->end() , std::less<OmexposurespecificproductRow>() );

	// OSW level products have an OSWID so use it but mak4e sure it is the numeric one.
	// We don't want degenerate filter names poluting our osw_id column.
	const Omsw* sw = pindex.omsw();
	for( Omsw::const_iterator CI = sw->begin()
	    ; CI != sw->end()
	    ; ++CI
	){
	    this->push_back( OmexposurespecificproductRow( row++
		, CI->instrument()
		, CI->exp_id()
		, CI->name()
		, CI->content()
		, CI->osw_id()
	    ));
	}
	std::sort(this->begin(), this->end() , std::less<OmexposurespecificproductRow>() );


    }

    void title( ostream& os ) const 
    {
	os  << "<tr><th class=\"string\">Filename</th><th class=\"string\">Content</th><th class=\"string\">Osw Id</th></tr>"
	    << "\n"
	;
    }

    friend ostream& operator<<( ostream& os , const Omexposurespecificproduct& r )
    {
	r.title(os);
	for( Omexposurespecificproduct::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }
    private:

};

class OmsourcesprocessedRow 
{
    public:
    OmsourcesprocessedRow( unsigned int v0
	, const Instrument& _Instrument
	, const Exp_id& _Exp_id
	, const Src_num& _Src_num
	, const Ra_obj& _Ra_obj
	, const Dec_obj& _Dec_obj
    )
    : _row(_row)
	, _instrument(_Instrument)
	, _exp_id(_Exp_id)
	, _src_num(_Src_num)
	, _ra_obj(_Ra_obj)
	, _dec_obj(_Dec_obj)
    {
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }
    Src_num src_num() const { return _src_num; }

    bool8 select( const Instrument& i , const Exp_id& e , const Src_num& s ) const
    {
	return ( _instrument == i && _exp_id == e && _src_num == s );
    }

    bool operator<( const OmsourcesprocessedRow& r ) const
    {
	// Sort order:  source number, instrument , source detection , exposure
	return  (_src_num < r._src_num)
	    || ( (_src_num == r._src_num) && (_instrument < r._instrument) ) 
	    || ( (_src_num == r._src_num) && (_instrument == r._instrument) && (_exp_id < r._exp_id) ) 
	    || ( (_src_num == r._src_num) && (_instrument == r._instrument) && (_exp_id == r._exp_id) && ( _src_num < r._src_num ) )
	;
    }

    std::string expanded_src_num() const
    {
	stringstream ss;
	ss << _src_num;
	ss << "&nbsp;(";
	_src_num.asHex(ss);
	ss << ")";

	return ss.str();
    }

    friend ostream& operator<<( ostream& os , const OmsourcesprocessedRow r )
    {
	std::string href1 = ppssumm_internal::makekey( r._instrument.pagename()  , r._instrument , r._exp_id , r._src_num );
	std::string href2 = ppssumm_internal::makekey( r._instrument.pagename()  , r._instrument , r._exp_id );
	std::ostream::fmtflags prev = os.setf( std::ios::fixed );
	unsigned int prec = os.precision();
	os  << "<tr>"
	    << "<td class=\"string\">" << "<a href=\"" << href1 << "\">" << r.expanded_src_num() << "</a></td>"
	    << "\n"
	    << "<td class=\"string\">" << r._instrument << "</td>"
	    << "\n"
	    << "<td class=\"string\">" << "<a href=\""<<  href2 << "\">" << r._exp_id << "</a></td>"
	    << "\n"
	    << "<td class=\"angle\">" << r._ra_obj.asHms() << "</td>"
	    << "\n"
	    << "<td class=\"angle\">" << r._dec_obj.asDms() << "</td>"
	    << "</tr>"
	    << "\n"
	;
	os.setf(prev);
	os.precision(prec);
        return os;
    }
    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Src_num _src_num;
    Ra_obj _ra_obj;
    Dec_obj _dec_obj;
   
};

class Omsourcesprocessed : public std::vector<OmsourcesprocessedRow>
{
    public:
    Omsourcesprocessed() : std::vector<OmsourcesprocessedRow>() {}
    Omsourcesprocessed( const Pindex& pindex ) : std::vector<OmsourcesprocessedRow>()
    {
	sbmap seen;
	const Omsrc* es = pindex.omsrc();
	unsigned int row=0;
	for( Omsrc::const_iterator CI = es->begin()
	    ; CI != es->end()
	    ; ++CI
	) {
	    std::string key = ppssumm_internal::makekey( CI->instrument() , CI->exp_id() , CI->src_num() );
	    if ( seen.find(key) == seen.end() )
	    {
		this->push_back ( OmsourcesprocessedRow( row++
		    , CI->instrument()
		    , CI->exp_id()
		    , CI->src_num()
		    , CI->ra_obj()
		    , CI->dec_obj()
		));
		seen[key] = true;
	    }
	}
	std::sort(this->begin(), this->end() , std::less<OmsourcesprocessedRow>() );
	
    }
    void title( ostream& os ) const
    {
	os  << "<tr>"
	    << "<th class=\"string\">Source</th>"
	    << "<th class=\"string\">Instrument</th>"
	    << "<th class=\"string\">Exposure</th>"
	    << "<th class=\"angle\">RA</th>"
	    << "<th class=\"angle\">Dec</th>"
	    << "</tr>"
	    << "\n"
	;
	os  << "<tr>"
	    << "<th class=\"string\" style=\" text-align: center; \">Dec (Hex)</th>"
	    << "<th class=\"string\" style=\" text-align: center; \">&nbsp</th>"
	    << "<th class=\"string\" style=\" text-align: center; \">&nbsp</th>"
	    << "<th class=\"angle\" style=\" text-align: center; \">[hms]</th>"
	    << "<th class=\"angle\" style=\" text-align: center; \">[&deg; &prime; &Prime;]</th>"
	    << "</tr>"
	    << "\n"
	;

    }

    friend ostream& operator<<( ostream& os , const Omsourcesprocessed r )
    {
	r.title(os);
	for( Omsourcesprocessed::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }
};

class OmsourcespecificproductRow
{
    public:
    OmsourcespecificproductRow( const unsigned int v0
	, const Instrument& v4
	, const Exp_id& v5
	, const Src_num& v6
	,const Name& v1
	,const Content& v2
	,const Osw_id& v3
    )
    : _row(v0)
    , _instrument(v4)
    , _exp_id(v5)
    , _src_num(v6)
    , _name(v1)
    , _content(v2)
    , _osw_id(v3)
    {
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }
    Src_num src_num() const { return _src_num; }

    bool8 select( const Instrument& i , const Exp_id& e , const Src_num& s ) const
    {
	return ( _instrument == i && _exp_id == e && _src_num == s );
    }
 
    bool operator<( const OmsourcespecificproductRow& r ) const
    {
	return  ( _src_num < r._src_num )
	    || ( ( _src_num == r._src_num ) &&  (_osw_id < r._osw_id) )
	    || ( ( _src_num == r._src_num ) &&  (_osw_id == r._osw_id) && ( _instrument < r._instrument ) )
	    || ( ( _src_num == r._src_num ) &&  (_osw_id == r._osw_id) && ( _instrument == r._instrument ) && ( _exp_id < r._exp_id ) )
	    || ( ( _src_num == r._src_num ) &&  (_osw_id == r._osw_id) && ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) &&  (_name < r._name))
	    || ( ( _src_num == r._src_num ) &&  (_osw_id == r._osw_id) && ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) 
		&&  (_name == r._name) && ( _content < r._content ))
	;
    }

    friend ostream& operator<<( ostream& os , const OmsourcespecificproductRow& r )
    {
	os  << "<tr>"
	    << "<td class=\"string\">" 
	    << r._name 
	    << "</td>"
	    << "<td class=\"string\">" 
	    << r._content 
	    << "</td>"
	    << "<td class=\"string\">" 
	    // We are only interested in numerical osw_id here.
	    << r._osw_id.osw_id()
	    << "</td>"
	    << "</tr>"
	    << "\n"
	;
	return os;
    }

    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Src_num _src_num;
    Name _name;
    Content _content;
    Osw_id _osw_id;
};

class Omsourcespecificproduct : public std::vector<OmsourcespecificproductRow>
{
    public:
    Omsourcespecificproduct() : std::vector<OmsourcespecificproductRow>() {}
    Omsourcespecificproduct( const Pindex& pindex ) : std::vector<OmsourcespecificproductRow>()
    {
	const Omsrc* es = pindex.omsrc();
	unsigned int row = 0;
	for( Omsrc::const_iterator CI = es->begin()
	    ; CI != es->end()
	    ; ++CI
	){
	    this->push_back( OmsourcespecificproductRow( row++
		, CI->instrument()
		, CI->exp_id()
		, CI->src_num()
		, CI->name()
		, CI->content()
		, CI->name().osw_id()
	    ));
	}
	
	std::sort(this->begin(), this->end() , std::less<OmsourcespecificproductRow>() );
    }

    void title( ostream& os ) const
    {
	os  << "<tr><th class=\"string\">Filename</th><th class=\"string\">Content</th><th class=\"string\">Osw Id</th></tr>"
	    << "\n"
	;
    }
    friend ostream& operator<<( ostream& os , const Omsourcespecificproduct r )
    {
	r.title(os);
	for( Omsourcespecificproduct::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }

    private:

};

#endif
