// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef PPSSUMM_PPSSUMMCOMMON
#define PPSSUMM_PPSSUMMCOMMON

#include <cctype>
#include <algorithm>
#include <functional>
#include <iterator>
#include <sstream>
#include "ppssumm_typedef.h"
#include "Instrument.h"
#include "Filename.h"
#include "Pindex.h"
#include "string_algo.h"
#include "FitsTable.h"
#include "ExtendFitsColumn.h"
#include "Datamode.h"
#include "STime.h"
#include "task_info.h"

#include "Dal.h"
#include "Tracer.h"
#include "errstr.h"

class PpssummCommon
{
public:
    /* 
     * This class wraps common data and functions to construct it
     * This data depends on the input files.
     */

    /*
     * fits summary file and product index. Needed by all pages.
     */

    typedef std::map< Instrument , bool8 > ibmap;
    typedef std::map< Instrument , bool8 >::const_iterator ibmap_ci;
    typedef std::map< Instrument , bool8 >::iterator ibmap_i;

    typedef std::map<std::string,Name> sfmap;
    typedef std::map<std::string,Name>::const_iterator sfmap_ci;
    typedef std::map<std::string,Name>::iterator sfmap_i;

    PpssummCommon( DataSet* fitsum , const std::string& pindex_file )
	: _taskversion(task_version()) , _page_id() 
    { 
	//_page_id gets passed around a bit so put iti in a stringlist
	std::string _page_id_init[] = 
	{
	     "obspage" 
	    ,"ppspage"
	    , "epxpage"
	    , "omxpage"
	    , "rgxpage"
	    , "catpage"
	    ,""
	};
	std::string* w = _page_id_init;
	while ( w->length() )
	{
	    _page_id.push_back( *w );
	    w++;
	}

	// Now collect lots of useful data
	extract_data( fitsum , pindex_file );
    } 
 
    // Getters
    std::string rev() const { return _rev; }
    std::string target() const { return _target; }
    std::string sched() const { return _sched; }
    std::string observer() const { return _observer; }
    std::string taskversion() const { return _taskversion; }
    std::string procdate() const { return _procdate; }
    std::string obs_id() const { return _obs_id; }
    std::string obs_start() const { return _obs_start; }
    std::string obs_stop() const { return _obs_stop; }
    std::string goname() const { return _goname; }
    std::string odf_version() const { return _odf_version; }

    stringlist page_id() const { return _page_id; }

    // Setters
    void rev( const std::string& s )  { _rev = s; }
    void target( const std::string& s )  { _target = s; }
    void sched( const std::string& s )  { _sched = s; }
    void observer( const std::string& s )  { _observer = s; }
    void taskversion( const std::string& s )  { _taskversion = s; }
    void procdate( const std::string& s )  { _procdate = s; }
    void obs_id( const std::string& s )  { _obs_id = s; }
    void obs_start( const std::string& s )  { _obs_start = s; }
    void obs_stop( const std::string& s )  { _obs_stop = s; }
    void goname( const std::string& s )  { _goname = s; }
    void odf_version( const std::string& s )  { _odf_version = s; }

    // Others
    std::string subtitle( const std::string& key ) const 
    {
	std::string r("");
	ssmap_ci q = _subtitle.find(key);
	if ( q != _subtitle.end() )
	{
	    r = q->second;
	} 
	return r;
    }

    //
    // Record of instruments processed
    //
    void set_instrument_active( const Instrument& i , bool8 proc = true )
    {
	// Delare the state of an instrument+Exp_id and whether it has been used in source detection
	_instrument_processed[i] = proc;	
    }

    bool8 was_instrument_active( const Instrument& i ) const
    {
	// Is the requested instrument+instrument pair listed in thre table ?
	bool8 r = false;
	ibmap_ci CI = _instrument_processed.find(i);
	r = (CI != _instrument_processed.end()) ? true : false ;
	return r;
    }

    //
    // Record of exposures processed
    //
    void set_exposure_processed( const Instrument& i , const Exp_id& e , bool8 srcdet = false)
    {
	// Delare the state of an instrument+Exp_id and whether it has been used in source detection
	std::string id = _inst_expid_key(i , e);
	_exposure_processed[id] = srcdet;	
    }

    bool8 was_exposure_processed( const Instrument& i , const Exp_id& e ) const 
    {
	// Is the requested instrument+exposure pair listed in thre table ?
	bool8 r = false;
	std::string id = _inst_expid_key(i , e);
	sbmap_ci CI = _exposure_processed.find(id);
	r = (CI != _exposure_processed.end()) ? true : false ;
	return r;
    }

    bool8 used_in_source_detection( const Instrument& i , const Exp_id& e ) const
    {
	// return the stored source detection value or false if not in the table.
	bool8 r = false;
	std::string id = _inst_expid_key(i , e);
	sbmap_ci CI = _exposure_processed.find(id);
	r = (CI != _exposure_processed.end()) ? CI->second : false ;
	return r;
    }

    //
    // Record for creating html links
    //

    void set_anchor( const Instrument& i , const Exp_id& e , Name f)
    {
	// Delare the state of an instrument+Exp_id and whether it has been used in source detection
	std::string id = _inst_expid_key(i , e);
	_html_anchor[id] = f;	
    }

    bool8 has_anchor( const Instrument& i , const Exp_id& e ) const
    {
	// Is the requested instrument+exposure pair listed in thre table ?
	bool8 r = false;
	std::string id = _inst_expid_key(i , e);
	sfmap_ci CI = _html_anchor.find(id);
	r = (CI != _html_anchor.end()) ? true : false ;
	return r;
    }

    std::string make_anchor( const Instrument& i , const Exp_id& e ) const
    {
	std::string id = _inst_expid_key(i , e);
	sfmap_ci CI = _html_anchor.find(id);
	Name f = (CI != _html_anchor.end()) ? CI->second : Name("") ;

	std::string r;
	r = f.asString() + "#" + id;
	return r;
    }
//

    private:

    std::string _rev;
    std::string _obs_id;
    std::string _target;
    std::string _sched;
    std::string _observer;
    std::string _taskversion;
    std::string _procdate;
    std::string _obs_start;
    std::string _obs_stop;
    std::string _goname;
    std::string _odf_version;
    stringlist _page_id;
    ssmap _subtitle;
    ibmap _instrument_processed;
    sbmap _exposure_processed;
    sfmap _html_anchor;

    PpssummCommon( const PpssummCommon& c ) {}
    void extract_data( DataSet*  , const std::string& );
    std::string _inst_expid_key( const Instrument& i , const Exp_id& e ) const
    {
	// Create an instrument+exposure Id key for use in maps.
	std::string id = i.asString() + "_" + e;
	return id;
    }


};

#endif
