// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef PPSSUMM_RGSINDEX
#define PPSSUMM_RGSINDEX

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <iterator>
#include <sstream>
#include <cmath>

// SAS 1
#include "Dal.h"
#include "Tracer.h"
#include "errstr.h"

// PPSSUMM Type classes
#include "ppssumm_typedef.h"
#include "string_algo.h"
#include "key_algo.h"
#include "PpssummCommon.h"
#include "Filename_typedef.h"
#include "Filename.h"
#include "Instrument.h"
#include "Src_num.h"
#include "Rgsorder.h"
#include "Degrees.h"
#include "Datamode.h"
#include "Pindex_typedef.h"
#include "Pindex.h"


class RgsmultiexposureproductRow
{
    public:
    RgsmultiexposureproductRow( const unsigned int v0 ,  const Name& v1 , const Content& v2 ) 
    : _row(v0) , _name(v1) , _content(v2) 
    {
    }

    bool operator<(const RgsmultiexposureproductRow& e) const
    {
	return _name < e.name();
    }

    friend ostream& operator<<( ostream& os , const RgsmultiexposureproductRow& r )
    {
	os  << "<tr>"
	    << "<td class=\"string\">" 
	    << r._name 
	    << "</td>"
	    << "<td class=\"string\">" 
	    << r._content 
	    << "</td>"
	    << "</tr>"
	    << "\n"
	;
	return os;
    }
    Name name() const { return _name; }
    Content content() const { return _content; }

    private:
    unsigned int _row;
    Name _name;
    Content _content;
};

class Rgsmultiexposureproduct : public std::vector<RgsmultiexposureproductRow>
{
    public:
    Rgsmultiexposureproduct() : std::vector<RgsmultiexposureproductRow>()  {}
    Rgsmultiexposureproduct(const Pindex& pindex) : std::vector<RgsmultiexposureproductRow>() 
    {
	const Rgsobs* rgsobs = pindex.rgsobs();
	unsigned long row=0;
	for ( Rgsobs::const_iterator CI = rgsobs->begin()
	    ; CI != rgsobs->end()
	    ; ++CI
	){
	    this->push_back( RgsmultiexposureproductRow( row++ , CI->name() , CI->content() ) );
	}

	std::sort(this->begin(), this->end() , std::less<RgsmultiexposureproductRow>() );
    }

    void title( ostream& os ) const
    {
	os  << "<tr><th class=\"string\">Filename</th><th class=\"string\">Content</th></tr>"
	    << "\n"
	;
    }
    friend ostream& operator<<( ostream& os , const Rgsmultiexposureproduct& r )
    {
	r.title(os);
	for( Rgsmultiexposureproduct::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }
    private:
};

class RgsexposuresprocessedRow
{
    public:
    RgsexposuresprocessedRow(
	const unsigned int v0
	,const Instrument& v1
	,const Exp_id& v2
	,const Scheduled& v3
	,const Mode& v4
	,const Datamode& v5
	,const Filtgrps& _Filtgrps
	,const Exp_duration& v8
	,const Exp_start& v10
	,const Exp_stop& v11
    ) : _row(v0)
	, _instrument(v1)
	, _exp_id(v2)
	, _scheduled(v3)
	, _mode(v4)
	, _datamode(v5)
	, _filtgrps(_Filtgrps)
	, _exp_duration(v8)
	, _exp_start(v10)
	, _exp_stop(v11)
    {
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }

    bool8 select( const Instrument& i , const Exp_id& e ) const
    {
	return ( _instrument == i && _exp_id == e );
    }

    bool operator<( const RgsexposuresprocessedRow& r ) const
    {
	return (_instrument < r._instrument)
	    || ((_instrument == r._instrument) && (_exp_start < r._exp_start))
	;
    }
    
    friend ostream& operator<<( ostream& os , const RgsexposuresprocessedRow& r )
    {
	ppssumm_internal::toString<bool8> tostring;

	std::string href1 = ppssumm_internal::makekey( r._instrument.pagename() , r._instrument , r._exp_id );
	std::string href2 = r._instrument.pagename() + "#sourcesprocessed";
	std::ostream::fmtflags incflags = os.setf( std::ios::fixed );
	unsigned int prec = os.precision();
	os << "<tr>"
	    << "<td class=\"string\">" << r._instrument << "</td>"
	    << "<td class=\"string\">" <<  "<a href=\"" << href1 << "\">" << r._exp_id << "</a></td>"
	    << "<td class=\"string\">" << tostring(r._scheduled) << "</td>"
	    << "<td class=\"string\">" << r._mode << "</td>"
	    << "<td class=\"string\">" << r._datamode << "</td>"
	;
	os.precision(0);
	os  << "<td class=\"string\">" << r._filtgrps << "</td>"
	    << "<td class=\"number\">" << r._exp_duration << "</td>"
	    << "<td class=\"date\">" << ppssumm_internal::truncTime(r._exp_start) << "</td>"
	    << "<td class=\"date\">" << ppssumm_internal::truncTime(r._exp_stop) << "</td>"
	    << "</tr>"
	    << "\n"
	;
	os.precision(prec);
	os.setf( incflags );
	return os;
    }

    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Scheduled _scheduled;
    Mode _mode;
    Datamode _datamode;
    Filtgrps _filtgrps;
    Exp_duration _exp_duration;
    Exp_start _exp_start;
    Exp_stop _exp_stop;
};

class Rgsexposuresprocessed : public std::vector<RgsexposuresprocessedRow>
{
public:
    Rgsexposuresprocessed() : std::vector<RgsexposuresprocessedRow>() {}
    Rgsexposuresprocessed( const Pindex& pindex ) : std::vector<RgsexposuresprocessedRow>() 
    {
	const Rgsexp* rgsexp = pindex.rgsexp();
	int row=0;
	for ( Rgsexp::const_iterator CI = rgsexp->begin()
	    ; CI != rgsexp->end()
	    ; ++CI
	){
	    std::string key = ppssumm_internal::makekey( CI->instrument() , CI->exp_id() );
	    sbmap_ci SI = _seen.find(key);
	    if ( SI == _seen.end() )
	    {
		_seen[key] = true;
		Exp_id exp_id = CI->exp_id();
		bool8 scheduled = false;
		if (! exp_id.empty() )
		{
		    scheduled = exp_id[0] == 'S';
		}
		this->push_back( RgsexposuresprocessedRow( row++
		    , CI->instrument() 
		    , CI->exp_id() 
		    , scheduled 
		    , CI->mode()
		    , CI->datamode()
		    , CI->filtgrps()
		    , CI->exp_duration()
		    , CI->exp_start()
		    , CI->exp_stop()
		));
	    }
	}

	std::sort(this->begin(), this->end() , std::less<RgsexposuresprocessedRow>() );
    }

    void title( ostream& os ) const
    {
	os << "<tr>"
	    << "<th class=\"string\">Inst.</th>"
	    << "<th class=\"string\">Exp.&nbsp;Id</th>"
	    << "<th class=\"string\">Sched</th>"
	    << "<th class=\"string\">Mode</th>"
	    << "<th class=\"string\">Datamode</th>"
	    << "<th class=\"string\">Event Filtering</th>"
	    << "<th class=\"number\">Duration</th>"
	    << "<th class=\"date\">Actual Start</th>"
	    << "<th class=\"date\">Actual End</th>"
	    << "</tr>"
	    << "\n"
	;
    }
    friend ostream& operator<<( ostream& os , const Rgsexposuresprocessed& r )
    {
	r.title(os);
	for( Rgsexposuresprocessed::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }
    const sbmap* seen() const { return &_seen; }
    private:
    sbmap _seen;
};

class RgsexposurespecificproductRow
{
    public:
    RgsexposurespecificproductRow( const unsigned int v0
	, const Instrument& v4
	, const Exp_id& v5
	,const Name& v1
	,const Content& v2
	,const Rgsorder& v3
    )
    : _row(v0)
    , _instrument(v4)
    , _exp_id(v5)
    , _name(v1)
    , _content(v2)
    , _rgsorder(v3)
    {

    }

    bool operator<( const RgsexposurespecificproductRow& r ) const
    {
	return ( _instrument < r._instrument )
	    || ( ( _instrument == r._instrument ) && ( _exp_id < r._exp_id ) )
	    || ( ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) && (_rgsorder < r._rgsorder) )
	    || ( ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) && (_rgsorder == r._rgsorder) && (_name < r._name))
	    || ( ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) && (_rgsorder == r._rgsorder) && (_name == r._name) && ( _content < r._content ))
	;
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }

    bool8 select( const Instrument& i , const Exp_id& e ) const
    {
	return ( _instrument == i && _exp_id == e );
    }

    friend ostream& operator<<( ostream& os , const RgsexposurespecificproductRow& r )
    {
	os  << "<tr><td class=\"string\">" 
	    << r._name 
	    << "</td>"
	    << "<td class=\"string\">" 
	    << r._content 
	    << "</td>"
	    << "<td class=\"number\">" 
	    << r._rgsorder
	    << "</td>"
	    << "\n"
	    << "</tr>"
	    << "\n"
	;
	return os;
    }

    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Name _name;
    Content _content;
    Rgsorder _rgsorder;
};

class Rgsexposurespecificproduct : public std::vector<RgsexposurespecificproductRow>
{
    public:
    Rgsexposurespecificproduct()
    : std::vector<RgsexposurespecificproductRow>()
    {
}

    Rgsexposurespecificproduct( const Pindex& pindex )
    : std::vector<RgsexposurespecificproductRow>()
    {
	const Rgsexp* ee = pindex.rgsexp();
	unsigned int row = 0;
	for( Rgsexp::const_iterator CI = ee->begin()
	    ; CI != ee->end()
	    ; ++CI
	){
    // rgsorder is currently missing from the PINDEX RGSEXP table.
    // Get it from the filename instead.
	    Name f(CI->name());
	    this->push_back( RgsexposurespecificproductRow( row++
		, CI->instrument()
		, CI->exp_id()
		, CI->name()
		, CI->content()
		, f.rgsorder()
	    ));
	}
	std::sort(this->begin(), this->end() , std::less<RgsexposurespecificproductRow>() );
    }

    void title( ostream& os ) const 
    {
	os  << "<tr><th class=\"string\">Filename</th><th class=\"string\">Content</th><th class=\"string\">Order</th></tr>"
	    << "\n"
	;
    }

    friend ostream& operator<<( ostream& os , const Rgsexposurespecificproduct& r )
    {
	r.title(os);
	for( Rgsexposurespecificproduct::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }
    private:

};

class RgssourcesprocessedRow 
{
    public:
    RgssourcesprocessedRow( unsigned int v0
	, const Instrument& _Instrument
	, const Exp_id& _Exp_id
	, const Src_num& _Src_num
	, const Ra_obj& _Ra_obj
	, const Dec_obj& _Dec_obj
	//, const Radec_err& _Radec_err
    )
    : _row(_row)
	, _instrument(_Instrument)
	, _exp_id(_Exp_id)
	, _src_num(_Src_num)
	, _ra_obj(_Ra_obj)
	, _dec_obj(_Dec_obj)
	//, _radec_err(_Radec_err)
    {
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }
    Src_num src_num() const { return _src_num; }

    bool8 select( const Instrument& i , const Exp_id& e , const Src_num& s ) const
    {
	return ( _instrument == i && _exp_id == e && _src_num == s );
    }

    bool operator<( const RgssourcesprocessedRow& r ) const
    {
	// Sort order:  source number, instrument , source detection , exposure
	return  (_src_num < r._src_num)
	    || ( (_src_num == r._src_num) && (_instrument < r._instrument) ) 
	    || ( (_src_num == r._src_num) && (_instrument == r._instrument) &&  ( _exp_id < r._exp_id ) ) 
	;
    }

    std::string expanded_src_num() const
    {
	stringstream ss;
	ss << _src_num;
	ss << "&nbsp;(";
	_src_num.asHex(ss);
	ss << ")";

	return ss.str();
    }

    friend ostream& operator<<( ostream& os , const RgssourcesprocessedRow r )
    {
	std::string href1 = ppssumm_internal::makekey( r._instrument.pagename() , r._instrument , r._exp_id , r._src_num );
	std::string href2 = ppssumm_internal::makekey( r._instrument.pagename() , r._instrument , r._exp_id );
	std::ostream::fmtflags prev = os.setf( std::ios::fixed );
	unsigned int prec = os.precision();
	os  << "<tr>"
	    << "<td class=\"string\">" << "<a href=\"" << href1 << "\">" << r.expanded_src_num() << "</a></td>"
	    << "\n"
	    << "<td class=\"string\">" << r._instrument << "</td>"
	    << "\n"
	    << "<td class=\"string\">" << "<a href=\""<<  href2 << "\">" << r._exp_id << "</a></td>"
	    << "\n"
	;
	Degrees d(0.0);
	d.setprec(1);
	os  << "<td class=\"angle\">" << r._ra_obj.asHms() << "</td>"
	    << "\n"
	;

	d.setprec(2);
	os  << "<td class=\"angle\">" << r._dec_obj.asDms() << "</td>"
	    << "\n"
	;

	//os  << "<td class=\"number\">" << r._radec_err.asAbrevDms() << "</td>"
	//    << "\n"
	//;

	os  << "</tr>"
	    << "\n"
	;
	os.setf(prev);
	os.precision(prec);
        return os;
    }
    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Src_num _src_num;
    Ra_obj _ra_obj;
    Dec_obj _dec_obj;
    //Radec_err _radec_err;
};

class Rgssourcesprocessed : public std::vector<RgssourcesprocessedRow>
{
    public:
    Rgssourcesprocessed() : std::vector<RgssourcesprocessedRow>() {}
    Rgssourcesprocessed( const Pindex& pindex ) : std::vector<RgssourcesprocessedRow>()
    {
	sbmap seen;
	const Rgssrc* es = pindex.rgssrc();
	unsigned int row=0;
	for( Rgssrc::const_iterator CI = es->begin()
	    ; CI != es->end()
	    ; ++CI
	) {
	    std::string key = ppssumm_internal::makekey( CI->instrument() , CI->exp_id() , CI->src_num() );
	    if ( seen.find(key) == seen.end() )
	    {
		this->push_back ( RgssourcesprocessedRow( row++
		    , CI->instrument()
		    , CI->exp_id()
		    , CI->src_num()
		    , CI->ra_obj()
		    , CI->dec_obj()
		   // , CI->radec_err()
		));
		seen[key] = true;
	    }
	}
	std::sort(this->begin(), this->end() , std::less<RgssourcesprocessedRow>() );
	
    }
    void title( ostream& os ) const
    {
	os  << "<tr>"
	    << "<th class=\"string\">Source</th>"
	    << "<th class=\"string\">Inst.</th>"
	    << "<th class=\"string\">Exp.&nbsp;Id</th>"
	    << "<th class=\"angle\">RA</th>"
	    << "<th class=\"angle\">Dec</th>"
	    //<< "<th class=\"angle\">RADec Err</th>"
	    << "</tr>"
	    << "\n"
	;
	os  << "<tr>"
	    << "<th class=\"unit\">Dec (Hex)</th>"
	    << "<th class=\"unit\">&nbsp</th>"
	    << "<th class=\"unit\">&nbsp</th>"
	    << "<th class=\"unit\">[hms]</th>"
	    << "<th class=\"unit\">[&deg; &prime; &Prime;]</th>"
	   // << "<th class=\"unit\">[arcsec]</th>"
	    << "</tr>"
	    << "\n"
	;
    }

    friend ostream& operator<<( ostream& os , const Rgssourcesprocessed r )
    {
	r.title(os);
	for( Rgssourcesprocessed::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }
};

class RgssourcespecificproductRow
{
    public:
    RgssourcespecificproductRow( const unsigned int v0
	, const Instrument& v4
	, const Exp_id& v5
	, const Src_num& v6
	,const Name& v1
	,const Content& v2
	,const Rgsorder& v3
    )
    : _row(v0)
    , _instrument(v4)
    , _exp_id(v5)
    , _src_num(v6)
    , _name(v1)
    , _content(v2)
    , _rgsorder(v3)
    {
    }

    Instrument instrument() const { return _instrument; }
    Exp_id exp_id() const { return _exp_id; }
    Src_num src_num() const { return _src_num; }

    bool8 select( const Instrument& i , const Exp_id& e , const Src_num& s ) const
    {
	return ( _instrument == i && _exp_id == e && _src_num == s );
    }
 
    bool operator<( const RgssourcespecificproductRow& r ) const
    {
	return  ( _src_num < r._src_num )
	    || ( ( _src_num == r._src_num ) &&  (_rgsorder < r._rgsorder) )
	    || ( ( _src_num == r._src_num ) &&  (_rgsorder == r._rgsorder) && ( _instrument < r._instrument ) )
	    || ( ( _src_num == r._src_num ) &&  (_rgsorder == r._rgsorder) && ( _instrument == r._instrument ) && ( _exp_id < r._exp_id ) )
	    || ( ( _src_num == r._src_num ) &&  (_rgsorder == r._rgsorder) && ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) &&  (_name < r._name))
	    || ( ( _src_num == r._src_num ) &&  (_rgsorder == r._rgsorder) && ( _instrument == r._instrument ) && ( _exp_id == r._exp_id ) 
		&&  (_name == r._name) && ( _content < r._content ))
	;
    }

    friend ostream& operator<<( ostream& os , const RgssourcespecificproductRow& r )
    {
	os  << "<tr>"
	    << "<td class=\"string\">" 
	    << r._name 
	    << "</td>"
	    << "<td class=\"string\">" 
	    << r._content 
	    << "</td>"
	    << "<td class=\"string\">" 
	    << r._rgsorder 
	    << "</td>"
	    << "</tr>"
	    << "\n"
	;
	return os;
    }

    private:
    unsigned int _row;
    Instrument _instrument;
    Exp_id _exp_id;
    Src_num _src_num;
    Name _name;
    Content _content;
    Rgsorder _rgsorder;
};

class Rgssourcespecificproduct : public std::vector<RgssourcespecificproductRow>
{
    public:
    Rgssourcespecificproduct() : std::vector<RgssourcespecificproductRow>() {}
    Rgssourcespecificproduct( const Pindex& pindex ) : std::vector<RgssourcespecificproductRow>()
    {
	const Rgssrc* es = pindex.rgssrc();
	unsigned int row = 0;
	for( Rgssrc::const_iterator CI = es->begin()
	    ; CI != es->end()
	    ; ++CI
	){
	    this->push_back( RgssourcespecificproductRow( row++
		, CI->instrument()
		, CI->exp_id()
		, CI->src_num()
		, CI->name()
		, CI->content()
		, CI->rgsorder()
	    ));
	}
	
	std::sort(this->begin(), this->end() , std::less<RgssourcespecificproductRow>() );
    }

    void title( ostream& os ) const
    {
	os  << "<tr><th class=\"string\">Filename</th><th class=\"string\">Content</th><th class=\"string\">Order</th></tr>"
	    << "\n"
	;
    }
    friend ostream& operator<<( ostream& os , const Rgssourcespecificproduct r )
    {
	r.title(os);
	for( Rgssourcespecificproduct::const_iterator CI = r.begin()
	    ; CI != r.end()
	    ; ++CI
	){
	    os << *CI;
	}
	return os;
    }

    private:

};

#endif
