// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef PPSSUMMS_SIMPLETABLE
#define PPSSUMMS_SIMPLETABLE

#include <string>
#include <vector>
#include <ostream>
#include "SAttribute.h"
#include "SColumn.h"
#include "stripws.h"
#include "Dal.h"

class STable : public std::vector<SColumn>
{
    // A container for tables
public:
    STable( const std::string& s="" ) 
    : name_(s) 
    {}

    std::string name() const { return name_; }
    STable& name(std::string n)
    { name_ = n ; return *this; }

    SColumn::size_type rows() const
    {
        /* return number of rows in longest column */
        SColumn::size_type rows = 0;
        for(const_iterator CI = begin()
            ; CI != end()
            ; CI++
        ) if (CI->size() > rows) rows =  CI->size();
        return rows;
    }

    void push_back(const SAttribute& sa) 
    {
        /*
         * SAttribute to Table Column
         * The SAttribute is taken as the name of a column.
         * If that column exists append the SAttribute value.
         * If it does not exist, create it using the SAttribute 
         * comment as the column comment and add the value to this
         * new column.
         */

        iterator i = begin();
        while ( (i != end()) && i->name() != sa.name()  ) i++;
        
        if ( i != end() ) {
            i->push_back(sa);
        } else {
                
            std::vector<SColumn>::push_back( SColumn(sa) );
        }
    }

    void push_back(const SColumn& sc) 
    {
        /*
         * If a column with name = sc.name() exists then append
         * sc to it.  Otherwise append sc as a new column to the table.
         */

        iterator i = begin();
        while ( (i != end()) && i->name() != sc.name()  ) i++;
        
        if ( i != end() ) {
            i->insert( sc.begin() , sc.end());
        } else {
            std::vector<SColumn>::push_back(sc);
        }
    }

    void insert(STable::const_iterator begin_ , STable::const_iterator end_) 
    {
        /* Add lots of columns */

        while (begin_ != end_) push_back(*begin_++); 
    }

    void pad( SColumn::size_type n=0 ) 
    {
        /* 
         * pad columns with up to n blank rows.
         * if n=0 then it will even up columns to be of 
         * uniform nr rows.
         */
        
        iterator i = begin();
        while ( i != end() ) {
            i->pad( (rows()+n) - i->size() );
            i++;
        }

    }

    friend std::ostream &operator<<(std::ostream& os , const STable& st)
    {
        os << "Table: " << st.name() << std::endl;
        os << "Name/Width/Rows/Comment" << std::endl;
        for (STable::const_iterator CI = st.begin()
            ; CI != st.end()
            ; CI++
        )  os << CI->name() << "/" << CI->width() << "/" << CI->size() << "/" << CI->comment() << std::endl;
        os << "***** END *****" << std::endl;
        return os;

    }
private:    
    std::string name_;
};

#endif
