// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include "sFunctions.h"

/* 
 * Functions for adding SAttributes , SColumns and STables
 * to Dal DataSet's and Table's
 */

//Add an attribute to the supplied dataset.
void addto(DataSet* ds , const SAttribute& satt) 
{
    if ( satt.name().length() && satt.value().length() ) {
        
        if ( ds->hasAttribute( satt.name() ) ) {
            
            Attribute* a = ds->attribute( satt.name8() );
            a->relabel( satt.comment() );
            *a = satt.value();

        } else {
          *ds->addAttribute( satt.name() , satt.comment() ) = satt.value();

        }
    }
}

//Add an attribute to the supplied table.
void addto(Table* table , const SAttribute& satt) 
{
    if (satt.name().length() && satt.value().length()) {

        if ( table->hasAttribute(satt.name()) ) {
            
            Attribute* a = table->attribute( satt.name8() );
            a->relabel(satt.comment());
            *a = satt.value();

        } else {
            *table->addAttribute( satt.name() , satt.comment() ) = satt.value();

        }
    }
}


void addto(Table* tbl , const SColumn& sc )
{
/*
 * Add a column to the supplied table.
 * I don't handle appending to existing columns.
 * This can be done before addto is called.
 * Adding to an existing column retains the charater width of the
 * existing column, truncating new entries if required
 */

    if ( tbl->hasColumn( sc.name() ) ) {

        /* I don't handle this case yet */
    } else {

        long diff = tbl->rows() - sc.size();
        if ( diff < 0 ) {
            tbl->insertRows( static_cast<unsigned long>( abs(diff) ));
        }

        std::vector<unsigned long> dim(1);
	// Minimum column width 1
        dim[0] = sc.width() ? static_cast<unsigned long>( sc.width() ) : 1;

        Column* col = tbl->addColumn(sc.name() , Column::DString , sc.comment() , "" , dim );
        ColumnData* coldat =  col->data();
        char* strdata = coldat->stringData();
        unsigned long rows = col->rows();
        unsigned int width= col->elements();

        unsigned long row = 0;
        
        /* 
         * values < width chars are padded with \0
         * values > width chars are truncated
         */
        for (SColumn::const_iterator CI = sc.begin()
            ;CI != sc.end() 
            ;CI++ , row++
        ) { 

            std::string s = *CI;
            
            if (s.length() > width)
                s = s.substr(0,width);
            else if (s.length() < width)
                s += std::string( width - s.length() , '\0');

            for( std::string::size_type i = 0 ; i < width; i++ ) strdata[ row*width+i ] = s[i];
        }

        /* Pad rows if added column has fewer rows than existing table */
        for( ; diff > 0 ; diff--) {
            for( std::string::size_type i = 0 ; i < width; i++ ) strdata[ (rows - diff)*width+i ] = '\0';
        }

        delete coldat;

    }
    
}

void addto(DataSet* ds , const STable& st )
{
    Table* tbl;
    if ( ds->hasBlock( st.name() ) ) {
        tbl = ds->table( st.name() );
    } else {

        tbl = ds->addTable( st.name() , static_cast<unsigned long>( st.rows() ) ) ; 
    }
    addto(tbl,st);

}

void addto(Table* tbl, const STable& st )
{
    for (STable::const_iterator CI = st.begin()
        ; CI != st.end()
        ;CI++
    ) {
        addto(tbl,*CI);    
    }

}

/* 
 * Turn summary file format line into an SAttribute.
 * Line formats translate as:
 * value => SAttribute("",value,"");
 * value / comment => SAttribute("",value,comment)
 * name = value // comment => SAttribute(name,value,comment)
 * alt allows an alternative name to be supplied, usefull as many
 * summary file keys are are > 8 char long.
 */

SAttribute MakeAttribute(const std::string& linein  , const std::string& alt ) {
    
    SAttribute a(alt);

    std::string line(stripws(linein));    

    if ( line.length() ) {
        
        std::string::size_type eq = line.find("=");
        std::string::size_type cm = line.find("/");

        if ( eq != std::string::npos && ! a.name().length() ) 
            a.name( stripws( line.substr(0 , eq ) ) );

        if ( eq != std::string::npos  ) { 
            a.value( stripws( line.substr( eq + 1 , (cm - eq - 1) ) ) );
        } else {
            a.value( stripws( line.substr( 0 , cm ) ) );
        }

        a.comment( stripws( line.substr( line.rfind("/")+1 , std::string::npos ) ) );

            
    }
    return a;
}

